/*
 * Copyright (c) 2011 Andrew Wells. All rights reserved.
 *
 *	Redistribution and use in source and binary forms, with or without modification, are
 *	permitted provided that the following conditions are met:
 *
 *		1. Redistributions of source code must retain the above copyright notice, this list of
 *     conditions and the following disclaimer.
 *
 *		2. Redistributions in binary form must reproduce the above copyright notice, this list
 *    of conditions and the following disclaimer in the documentation and/or other materials
 *    provided with the distribution.
 *
 *	THIS SOFTWARE IS PROVIDED BY Andrew Wells ''AS IS'' AND ANY EXPRESS OR IMPLIED
 *	WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *	FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL Andrew Wells OR
 *	CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *	CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 *	SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *	ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *	NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 *	ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * 	The views and conclusions contained in the software and documentation are those of the
 *	authors and should not be interpreted as representing official policies, either expressed
 *	or implied, of Andrew Wells.
 */

package aw.Mallorn.tree;

import java.util.ArrayList;
import java.util.LinkedList;

public class KDTree {
	KDTreeBucket trunk;
	final static int defaultMaxBucketCapacity = 24;
	int dimensions;
	LinkedList<double[]> listOfPointsInOrderAdded;
	int maxSize;

	public KDTree(int dimensions) {
		trunk = new KDTreeBucket(defaultMaxBucketCapacity, dimensions);
		this.dimensions = dimensions;
		listOfPointsInOrderAdded = null;
	}

	public KDTree(int maxBucketCapacity, int dimensions) {
		trunk = new KDTreeBucket(maxBucketCapacity, dimensions);
		this.dimensions = dimensions;
		listOfPointsInOrderAdded = null;
	}

	public KDTree(int maxBucketCapacity, int dimensions, int maxSize) {
		trunk = new KDTreeBucket(maxBucketCapacity, dimensions);
		this.dimensions = dimensions;
		listOfPointsInOrderAdded = new LinkedList<double[]>();
		this.maxSize = maxSize;
	}

	/**
	 * @param point   The point to add.
	 * @return	In a size limited tree, this returns the removed point.
	 */
	public double[] addPoint(double[] point) {
		double[] returnPoint = null;
		if (listOfPointsInOrderAdded != null) {
			listOfPointsInOrderAdded.addLast(point);
			if (listOfPointsInOrderAdded.size() > maxSize) {
				returnPoint = listOfPointsInOrderAdded.removeFirst();
				removePoint(returnPoint);
			}
		}

		KDTreeBucket currentBucket = trunk;

		while (!currentBucket.isTerminal) {
			extendBounds(point, currentBucket);
			if (point[currentBucket.dividingDimension] < currentBucket.divider)
				currentBucket = currentBucket.LeftBucket;
			else
				currentBucket = currentBucket.RightBucket;
		}
		extendBounds(point, currentBucket);
		currentBucket.points[currentBucket.currentLastPoint++] = point;
		if (currentBucket.currentLastPoint == currentBucket.maxSize) {
			setDivider(currentBucket);
			// setting Left and Right buckets, isTerminal and calling
			// redistribute points will be done in setDivider() so that they are
			// not done if no divider is set. This happens when the points
			// in the bucket are all identical.
			// redistribute points also sets the points list to null.
		}
		return returnPoint;
	}

	private void removePoint(double[] point) {
		KDTreeBucket currentBucket = trunk;

		while (!currentBucket.isTerminal) {
			if (point[currentBucket.dividingDimension] < currentBucket.divider)
				currentBucket = currentBucket.LeftBucket;
			else
				currentBucket = currentBucket.RightBucket;
		}
		int indexOfPoint = -1;
		for (int i = 0; i < currentBucket.points.length; i++) {
			if (currentBucket.points[i] == (point)) {
				indexOfPoint = i;
				break;
			}
		}
		if (currentBucket.currentLastPoint > 1) {
			currentBucket.points[indexOfPoint] = currentBucket.points[--currentBucket.currentLastPoint];

			ArrayList<Integer> effectsDimensionMBR = new ArrayList<Integer>();
			double[][] newMBR = new double[dimensions][2];
			for (int i = 0; i < dimensions; i++) {
				double max = currentBucket.points[0][i];
				double min = currentBucket.points[0][i];

				for (int j = 1; j < currentBucket.currentLastPoint; j++) {
					if (currentBucket.points[j][i] > max) {
						max = currentBucket.points[j][i];
					} else if (currentBucket.points[j][i] < min) {
						min = currentBucket.points[j][i];
					}
				}
				newMBR[i][0] = min;
				newMBR[i][1] = max;
				if (newMBR[i][0] != currentBucket.minimumBoundingRectangle[i][0]
						|| newMBR[i][1] != currentBucket.minimumBoundingRectangle[i][1])
					effectsDimensionMBR.add(i);
			}

				currentBucket = currentBucket.Parent;
				while (currentBucket != null && effectsDimensionMBR.size() > 0) {
					ArrayList<Integer> newEffectsDimensionMBR = new ArrayList<Integer>();
					for (Integer dimension : effectsDimensionMBR) {
						double min = newMBR[dimension][0];
						double max = newMBR[dimension][1];
						if (currentBucket.RightBucket.minimumBoundingRectangle[dimension][0] < min) {
							min = currentBucket.RightBucket.minimumBoundingRectangle[dimension][0];
							newEffectsDimensionMBR.add(dimension);
						} else if (currentBucket.RightBucket.minimumBoundingRectangle[dimension][1] > max) {
							max = currentBucket.RightBucket.minimumBoundingRectangle[dimension][1];
							newEffectsDimensionMBR.add(dimension);
						}
						if (currentBucket.LeftBucket.minimumBoundingRectangle[dimension][0] < min) {
							min = currentBucket.LeftBucket.minimumBoundingRectangle[dimension][0];
							newEffectsDimensionMBR.add(dimension);
						} else if (currentBucket.LeftBucket.minimumBoundingRectangle[dimension][1] > max) {
							max = currentBucket.LeftBucket.minimumBoundingRectangle[dimension][1];
							newEffectsDimensionMBR.add(dimension);
						}
						currentBucket.minimumBoundingRectangle[dimension][0] = min;
						currentBucket.minimumBoundingRectangle[dimension][1] = max;
					}
					effectsDimensionMBR = newEffectsDimensionMBR;
					currentBucket = currentBucket.Parent;
				}
			
		} else {
			while (currentBucket.Parent != null) {
				if (currentBucket.Parent.RightBucket == currentBucket) {
					if (currentBucket.Parent.LeftBucket.isTerminal
							&& currentBucket.Parent.LeftBucket.currentLastPoint == 0) {
							currentBucket = currentBucket.Parent;
					} else {
						currentBucket = new KDTreeBucket(currentBucket.Parent);
						currentBucket.Parent.RightBucket = currentBucket;
						currentBucket.Parent.minimumBoundingRectangle = currentBucket.Parent.LeftBucket.minimumBoundingRectangle;
						break;
					}
				} else {
					if (currentBucket.Parent.RightBucket.isTerminal
							&& currentBucket.Parent.RightBucket.currentLastPoint == 0) {
							currentBucket = currentBucket.Parent;
					} else {
						currentBucket = new KDTreeBucket(currentBucket.Parent);
						currentBucket.Parent.LeftBucket = currentBucket;
						currentBucket.Parent.minimumBoundingRectangle = currentBucket.Parent.RightBucket.minimumBoundingRectangle;
						break;
					}
				}
			}
			if (currentBucket.Parent == null) {
				currentBucket = new KDTreeBucket(null);
			}
		}
	}

	// TODO multiple implementations (in C++ version?)
	private void setDivider(KDTreeBucket bucket) {
		double widestVariance = 0;
		double max, min;
		max = 0;
		min = 0;
		double bestDimensionMin = 0;
		int dimensionWithWidestVariance = 0;

		for (int j = 0; j < dimensions; j++) {
			double[] currentPoint = bucket.points[0];
			max = currentPoint[j];
			min = currentPoint[j];
			for (int i = 1; i < bucket.maxSize; i++) {
				currentPoint = bucket.points[i];
				if (currentPoint[j] < min)
					min = currentPoint[j];
				else if (currentPoint[j] > max)
					max = currentPoint[j];
			}
			if (max - min > widestVariance) {
				widestVariance = max - min;
				dimensionWithWidestVariance = j;
				bestDimensionMin = min;
			}
		}

		// TODO Use median of widest dimension instead?
		// if there is no variance double the bucket size
		// this will and should carry over to child nodes since they will
		// have these points in them.
		if (widestVariance == 0) {
			bucket.maxSize *= 2;
			double[][] tempPoints = bucket.points;
			bucket.points = new double[bucket.maxSize][dimensions];
			System.arraycopy(tempPoints, 0, bucket.points, 0, tempPoints.length);
			tempPoints = null;
		} else {
			bucket.LeftBucket = new KDTreeBucket(bucket);
			bucket.RightBucket = new KDTreeBucket(bucket);
			bucket.dividingDimension = dimensionWithWidestVariance;
			bucket.divider = bestDimensionMin + (widestVariance) * 0.5;
			redistributePoints(bucket);
			bucket.points = null;
			bucket.isTerminal = false;
		}
	}

	private void redistributePoints(KDTreeBucket bucket) {
		for (int i = 0; i < bucket.maxSize; i++) {
			double[] currentPoint = bucket.points[i];
			if (currentPoint[bucket.dividingDimension] < bucket.divider) {
				bucket.LeftBucket.points[bucket.LeftBucket.currentLastPoint++] = (currentPoint);
				extendBounds(currentPoint, bucket.LeftBucket);
			} else {
				bucket.RightBucket.points[bucket.RightBucket.currentLastPoint++] = (currentPoint);
				extendBounds(currentPoint, bucket.RightBucket);
			}
		}
	}

	private void extendBounds(double[] point, KDTreeBucket bucket) {
		for (int i = 0; i < dimensions; i++) {
			if (point[i] < bucket.minimumBoundingRectangle[i][0])
				bucket.minimumBoundingRectangle[i][0] = point[i];
			if (point[i] > bucket.minimumBoundingRectangle[i][1])
				bucket.minimumBoundingRectangle[i][1] = point[i];
		}
	}

	/**
	 * 
	 * @param point		Query point for N Nearest Neighbors search.
	 * @param n		Number of neighbors to return.  This should never be 0.
	 * @return	The N nearest neighbors of the point.
	 */
	public double[][] getNNearestPoints(double[] point, int n) {
		// System.out.println("get points started");

		// TODO use a heap

		double[][] tempNearestPoints = new double[n][];
		// double[][] nearestPoints = new double[n][]; // this will be a binary
		// max heap
		double[] nearestPointsDistances = new double[n];
		for (int i = 0; i < n; i++) {
			nearestPointsDistances[i] = Double.POSITIVE_INFINITY;
		}
		int currentNearestPointsIndex = 0;
		KDTreeBucket currentBucket = trunk;
		double distToWorstNearPoint = Double.POSITIVE_INFINITY; // just using a large
															// number

		// TODO optimize more this version may be worse for robocode

		// use square of actual distances:
		do {
			if (currentBucket.isTerminal) {
				for (int i = 0; i < currentBucket.currentLastPoint; i++) {
					double currentDist = squareDistBetweenPoints(
							currentBucket.points[i], point);
					if (currentDist <= distToWorstNearPoint) {

						// // if the heap is full, remove the furthestPoint and
						// // set heap
						// if(currentNearestPointIndex == n) {
						// nearestPoints[0] = (nearestPointsDistances[1] -
						// nearestPointsDistances[2] > 0) ? nearestPoints[1] :
						// nearestPoints[2];
						// }
						// // insert new point
						// nearestPoints[currentNearestPointIndex++] =
						// currentBucket.points[i];
						// // check that new heap is valid
						//
						// // set distToWorstNearPoint

						tempNearestPoints[currentNearestPointsIndex] = currentBucket.points[i];
						nearestPointsDistances[currentNearestPointsIndex] = currentDist;
						for (int j = currentNearestPointsIndex - 1; j > -1
								&& currentDist <= nearestPointsDistances[j]; j--) {
							double[] temp;
							double tempDist;
							temp = tempNearestPoints[j];
							tempDist = nearestPointsDistances[j];
							tempNearestPoints[j] = tempNearestPoints[j + 1];
							nearestPointsDistances[j] = nearestPointsDistances[j + 1];
							tempNearestPoints[j + 1] = temp;
							nearestPointsDistances[j + 1] = tempDist;
						}
						if (currentNearestPointsIndex < n - 1)
							currentNearestPointsIndex++;
						distToWorstNearPoint = nearestPointsDistances[currentNearestPointsIndex];

					}
				}
				// if we can guarantee that this is not null, we can save time:
				currentBucket = currentBucket.Parent;
			}

			else if (currentBucket.nodesVisited == NodesVisited.None) {
//				TODO use MBR
				if (point[currentBucket.dividingDimension] < currentBucket.divider) {
					currentBucket.nodesVisited = NodesVisited.Left;
					if (distToWorstNearPoint >= squareDistFromPointToHyperRactangle(
							point, currentBucket.LeftBucket))
						currentBucket = currentBucket.LeftBucket;
				} else {
					currentBucket.nodesVisited = NodesVisited.Right;
					if (distToWorstNearPoint >= squareDistFromPointToHyperRactangle(
							point, currentBucket.RightBucket))
						currentBucket = currentBucket.RightBucket;
				}
			} else if (currentBucket.nodesVisited == NodesVisited.Right
					&& distToWorstNearPoint >= squareDistFromPointToHyperRactangle(
							point, currentBucket.LeftBucket)) {
				currentBucket.nodesVisited = NodesVisited.All;
				currentBucket = currentBucket.LeftBucket;
			} else if (currentBucket.nodesVisited == NodesVisited.Left
					&& distToWorstNearPoint >= squareDistFromPointToHyperRactangle(
							point, currentBucket.RightBucket)) {
				currentBucket.nodesVisited = NodesVisited.All;
				currentBucket = currentBucket.RightBucket;
			} else {
				currentBucket.nodesVisited = NodesVisited.None;
				currentBucket = currentBucket.Parent;
			}

		} while (currentBucket != null);

		return tempNearestPoints;
	}

	public static double squareDistBetweenPoints(double[] pointOne,
			double[] pointTwo) {
		double squareDist = 0.0;
		for (int i = 0; i < pointOne.length; i++) {
			double factor = pointOne[i] - pointTwo[i];
			// if the points will lie at exactly the same spot
			// on some axes frequently enough, this could be:

			// if(factor != 0)
			squareDist += factor * factor;
		}
		return squareDist;
	}

	private double squareDistFromPointToHyperRactangle(
			double[] pointCoordinates, KDTreeBucket bucket) {
		double factor;
		double squareDist = 0.0;
		for (int i = 0; i < dimensions; i++) {
			if (pointCoordinates[i] < bucket.minimumBoundingRectangle[i][0]) {
				factor = bucket.minimumBoundingRectangle[i][0]
						- pointCoordinates[i];
				squareDist += factor * factor;
			} else if (pointCoordinates[i] > bucket.minimumBoundingRectangle[i][1]) {
				factor = pointCoordinates[i]
						- bucket.minimumBoundingRectangle[i][1];
				squareDist += factor * factor;
			}
		}
		return squareDist;
	}

/**
 * 
 * @param point  The point to perform the N Nearest Neighbors search for.
 * @param n		The number of points to return, this should never be 0;
 * @param weights	The weights for each dimension.
 * @return	The N Nearest Neighbors for the query point.
 */
	public double[][] getNNearestPoints(double[] point, int n, double[] weights) {
		// System.out.println("get points started");

		// TODO use a heap

		double[][] tempNearestPoints = new double[n][];
		// double[][] nearestPoints = new double[n][]; // this will be a binary
		// max heap
		double[] nearestPointsDistances = new double[n];
		for (int i = 0; i < n; i++) {
			nearestPointsDistances[i] = Double.POSITIVE_INFINITY;
		}
		int currentNearestPointsIndex = 0;
		KDTreeBucket currentBucket = trunk;
		double distToWorstNearPoint = Double.POSITIVE_INFINITY;

		// TODO optimize more this version may be worse for robocode

		// use square of actual distances:
		do {
			if (currentBucket.isTerminal) {
				for (int i = 0; i < currentBucket.currentLastPoint; i++) {
					double currentDist = squareDistBetweenPoints(
							currentBucket.points[i], point, weights);
					if (currentDist <= distToWorstNearPoint) {

						// // if the heap is full, remove the furthestPoint and
						// // set heap
						// if(currentNearestPointIndex == n) {
						// nearestPoints[0] = (nearestPointsDistances[1] -
						// nearestPointsDistances[2] > 0) ? nearestPoints[1] :
						// nearestPoints[2];
						// }
						// // insert new point
						// nearestPoints[currentNearestPointIndex++] =
						// currentBucket.points[i];
						// // check that new heap is valid
						//
						// // set distToWorstNearPoint

						tempNearestPoints[currentNearestPointsIndex] = currentBucket.points[i];
						nearestPointsDistances[currentNearestPointsIndex] = currentDist;
						for (int j = currentNearestPointsIndex - 1; j > -1
								&& currentDist <= nearestPointsDistances[j]; j--) {
							double[] temp;
							double tempDist;
							temp = tempNearestPoints[j];
							tempDist = nearestPointsDistances[j];
							tempNearestPoints[j] = tempNearestPoints[j + 1];
							nearestPointsDistances[j] = nearestPointsDistances[j + 1];
							tempNearestPoints[j + 1] = temp;
							nearestPointsDistances[j + 1] = tempDist;
						}
						if (currentNearestPointsIndex < n - 1)
							currentNearestPointsIndex++;
						distToWorstNearPoint = nearestPointsDistances[currentNearestPointsIndex];

					}
				}
				// if we can guarantee that this is not null, we can save time:
				currentBucket = currentBucket.Parent;
			}

			else if (currentBucket.nodesVisited == NodesVisited.None) {
				if (point[currentBucket.dividingDimension] < currentBucket.divider) {
					currentBucket.nodesVisited = NodesVisited.Left;
					if (distToWorstNearPoint >= squareDistFromPointToHyperRactangle(
							point, currentBucket.LeftBucket, weights))
						currentBucket = currentBucket.LeftBucket;
				} else {
					currentBucket.nodesVisited = NodesVisited.Right;
					if (distToWorstNearPoint >= squareDistFromPointToHyperRactangle(
							point, currentBucket.RightBucket, weights))
						currentBucket = currentBucket.RightBucket;
				}
			} else if (currentBucket.nodesVisited == NodesVisited.Right
					&& distToWorstNearPoint >= squareDistFromPointToHyperRactangle(
							point, currentBucket.LeftBucket, weights)) {
				currentBucket.nodesVisited = NodesVisited.All;
				currentBucket = currentBucket.LeftBucket;
			} else if (currentBucket.nodesVisited == NodesVisited.Left
					&& distToWorstNearPoint >= squareDistFromPointToHyperRactangle(
							point, currentBucket.RightBucket, weights)) {
				currentBucket.nodesVisited = NodesVisited.All;
				currentBucket = currentBucket.RightBucket;
			} else {
				currentBucket.nodesVisited = NodesVisited.None;
				currentBucket = currentBucket.Parent;
			}

		} while (currentBucket != null);

		return tempNearestPoints;
	}

	public static double squareDistBetweenPoints(double[] pointOne,
			double[] pointTwo, double[] weights) {
		double squareDist = 0.0;
		for (int i = 0; i < pointOne.length; i++) {
			double factor = (pointOne[i] - pointTwo[i]) * weights[i];
			// if the points will lie at exactly the same spot
			// on some axes frequently enough, this could be:

			// if(factor != 0)
			squareDist += factor * factor;
		}
		return squareDist;
	}

	private double squareDistFromPointToHyperRactangle(
			double[] pointCoordinates, KDTreeBucket bucket, double[] weights) {
		double factor;
		double squareDist = 0.0;
		for (int i = 0; i < dimensions; i++) {
			if (pointCoordinates[i] < bucket.minimumBoundingRectangle[i][0]) {
				factor = (bucket.minimumBoundingRectangle[i][0] - pointCoordinates[i])
						* weights[i];
				squareDist += factor * factor;
			} else if (pointCoordinates[i] > bucket.minimumBoundingRectangle[i][1]) {
				factor = (pointCoordinates[i] - bucket.minimumBoundingRectangle[i][1])
						* weights[i];
				squareDist += factor * factor;
			}
		}
		return squareDist;
	}

}
