/**
 * Copyright (c) 2011 Chase
 * 
 * This software is provided 'as-is', without any express or implied
 * warranty. In no event will the authors be held liable for any damages
 * arising from the use of this software.
 * 
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, and to alter it and redistribute it
 * freely, subject to the following restrictions:
 * 
 *    1. The origin of this software must not be misrepresented; you must not
 *    claim that you wrote the original software. If you use this software
 *    in a product, an acknowledgment in the product documentation would be
 *    appreciated but is not required.
 * 
 *    2. Altered source versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 * 
 *    3. This notice may not be removed or altered from any source
 *    distribution.
 */
/*
 * Dear Future Chase,
 * 
 * I cannot even begin to express how deeply sorry I am.
 * But I assure you, at the time, I (probably) knew what I was doing.
 * 
 * -Chase
 * 
 * P.S. I am trying to leave comments,
 *      but you know how bad I am at those.
 */

package cs.move;

import static cs.utils.FastMath.*;

import java.awt.Color;
import java.awt.geom.Line2D;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import cs.Chikaku;
import cs.geom.*;
import cs.utils.*;

import ags.utils.KdTree;
import ags.utils.KdTree.Entry;
import robocode.*;
import robocode.util.Utils;

/**
 * Movement
 * <br/><br/>
 * In Japanese, Kakeru means "Dash" (in some cases)
 * @author Chase
 */
public abstract class Kakeru extends Chikaku {
	/**
	 * Keep in mind with heat waves we will have waves 2 rounds earlier
	 * the the gun heat would otherwise indicate
	 */
	private static final int TimeBeforeFirstWaveToStartTurning = 4;
	private static final int MaxTurnHistoryLength = 32;
	private static final int MaxHeatWaveQueueLength = 4;
	private static final int AverageSingleRoundTime = 1200;

	private static final boolean DEBUG_WAVELESS_MOVEMENT = false;
	private static final boolean DEBUG_BULLET_HANDLING = true;

	/*
	 * Distancer Constants, control how the distancer acts,
	 * limiter is linear equation
	 */
	private static final double distBestDistance = 500.0;

	/* distancer angle limiting */
	private static final double distMinDistance = 300;
	private static final double distMinRotation = Math.PI/4.0;
	private static final double distMaxDistance = 800;
	private static final double distMaxRotation = Math.PI/10.0;
	private static final double distMultiplier = (distMaxRotation-distMinRotation)/(distMaxDistance-distMinDistance);
	private static final double distAdder = distMinRotation-distMultiplier*distMinDistance;

	private static final MaximumDeque<Double> enemyBulletPower;
	private static final KdTree<Data> tree;

	static {
		enemyBulletPower = new MaximumDeque<Double>(MaxHeatWaveQueueLength);
		enemyBulletPower.addLast(3.0);

		int dimensions = new Data().getDataArray().length;
		tree = new KdTree.SqrEuclid<Data>(dimensions, 0);

		DuelSeed.seed(tree);
	}

	private final LinkedList<Wave<Data>> enemyWaves;
	private final MaximumDeque<State> history;
	private Vector lastEnemyPosition;
	private double lastEnemyEnergy;
	private double lastEnemyVelocity;
	private long timeSinceDirectionChange;
	private double enemyGunHeatx10;

	private int orbitDirection = 1;

	private Wave<Data> lastWave = null;
	private long lastWaveFireTime = 0;

	private static int enemyBulletFired = 0;
	private static int enemyBulletIntercepted = 0;
	private static int enemyBulletHit = 0;

	public Kakeru() {
		enemyWaves = new LinkedList<Wave<Data>>();
		history = new MaximumDeque<State>(MaxTurnHistoryLength);
		lastEnemyPosition = new Vector(-1,-1);
		lastEnemyEnergy = 100;
	}

	@Override
	public void onRoundStarted(Event e) {
		super.onRoundStarted(e);

		/* Set it to be the same as our gun heat on round start! */
		enemyGunHeatx10 = status.getGunHeat() * 10;

		/* Our scans are 1 round late */
		enemyGunHeatx10 -= coolingRate*10;
	}

	private static void prepareWaveData(Wave<Data> wave) {
		wave.data.bulletFlightTime = wave.data.distance / wave.speed;
	}

	@Override
	public void onScannedRobot(ScannedRobotEvent e) {
		super.onScannedRobot(e);

		double angleToEnemy = e.getBearingRadians() + status.getHeadingRadians();
		double enemyVelocity = e.getVelocity();

		State state = new State();

		state.myPosition = myPosition;
		state.enemyPosition = myPosition.projectNew(angleToEnemy, e.getDistance());
		state.data.lateralVelocity = status.getVelocity()*Math.sin(e.getBearingRadians());
		state.data.advancingVelcity = status.getVelocity()*Math.cos(e.getBearingRadians());
		state.data.distance = e.getDistance();
		state.data.globalTime = globalTime;

		int direction = sign(state.data.lateralVelocity);
		if(direction == orbitDirection) {
			++timeSinceDirectionChange;
		} else {
			timeSinceDirectionChange = 0;
		}
		orbitDirection = direction;

		history.addFirst(state);

		state.data.timeSinceDirectionChange = timeSinceDirectionChange;

		state.data.wallForward = Tools.getWallDistance(
				state.enemyPosition,
				field,
				state.data.distance,
				state.enemyPosition.angleTo(state.myPosition),
				direction);

		state.data.wallReverse = Tools.getWallDistance(
				state.enemyPosition,
				field,
				state.data.distance,
				state.enemyPosition.angleTo(state.myPosition),
				-direction);

		state.data.distLast10 = myPosition.distance(history.get(min(10,history.size()-1)).myPosition);

		if(abs(enemyVelocity) == 0.0) {
			if(abs(lastEnemyVelocity) > 2.0) {
				//Hit the wall!
				lastEnemyEnergy -= Rules.getWallHitDamage(abs(lastEnemyVelocity));
			} else {
				//TODO slightly more complicated check near wall and heading check
			}
		}


		double enemyEnergy = e.getEnergy();
		/*
		 * Fire new imaginary wave if we can, heat waves
		 * are always 2 ticks previous to a real one 'possibly'
		 * being fired
		 */
		if(enemyGunHeatx10 <= 20*coolingRate && enemyGunHeatx10 > 10*coolingRate) {
			double power = 0;
			for(Double d : enemyBulletPower)
				power += d;
			power /= enemyBulletPower.size();

			power = min(enemyEnergy,power);

			/* Create the Wave */
			Wave<Data> wave = createMovementWave(power,direction);
			wave.setLocation(simulatedEnemyPosition);
			wave.imaginary = true;

			/*
			 * Only +1 because real waves are -1 in time,
			 * meaning that by the time we detect them, it has
			 * already been 1 turn since it was fired, meaning
			 * heat waves +2 makes it only +1 here. :-)
			 */
			wave.fireTime = time + 1;
			wave.directAngle = wave.angleTo(state.myPosition);

			wave.data = state.data;
			prepareWaveData(wave);

			enemyWaves.add(wave);
		}

		/* Check for new enemy wave */
		double energyDelta = lastEnemyEnergy - enemyEnergy;
		if(energyDelta > 0 && energyDelta <= 3.0) {
			enemyGunHeatx10 = Rules.getGunHeat(energyDelta)*10;

			State old = history.get(2);
			direction = sign(old.data.lateralVelocity);

			/* Create the Wave */
			Wave<Data> wave = createMovementWave(energyDelta,direction);
			wave.setLocation(lastEnemyPosition);

			wave.fireTime = time - 1;
			wave.directAngle = wave.angleTo(old.myPosition);

			wave.data = old.data;

			enemyBulletPower.addLast(wave.power);

			boolean add = true;

			if(Math.abs(energyDelta-0.1) < 0.001) {
				/* anti inactivity counter */
				if(time - lastWaveFireTime == 1) {
					enemyWaves.remove(lastWave);
					add = false;
					--enemyBulletFired;
				}

				lastWave = wave;
				lastWaveFireTime = time;
			}

			if(add) {
				prepareWaveData(wave);
				calculateShadowsForWave(wave);
				enemyWaves.add(wave);
				++enemyBulletFired;
			}
		}

		if(DEBUG_WAVELESS_MOVEMENT) {
			if(lastEnemyPosition.x < 0) {
				lastEnemyPosition = myPosition.projectNew(angleToEnemy, 50);
			}
		} else {
			lastEnemyPosition = state.enemyPosition;
		}


		lastEnemyVelocity = enemyVelocity;
		lastEnemyEnergy = enemyEnergy;

	}

	private Wave<Data> createMovementWave(double power, int direction) {
		Wave<Data> wave = new Wave<Data>();
		wave.power = power;
		wave.speed = Rules.getBulletSpeed(wave.power);
		wave.escapeAngle = Math.asin(8.0/wave.speed) * direction;
		return wave;
	}

	@Override
	public void onTurnEnded(Event e) {
		super.onTurnEnded(e);

		/* Don't move if we are scanning, this is mostly start
		 * of round fast radar finding protection
		 */
		if(initialTick != 0)
			return;

		/* We track the gun heat because it is important
		 * for our heat/imaginary waves
		 */
		enemyGunHeatx10 -= coolingRate*10;
		if(enemyGunHeatx10 <= 0)
			enemyGunHeatx10 = 0;

		/* Take are of all waves that cannot possibly even conceivably hit us anymore */
		checkWaves();

		/* Run our 'actual' movement */
		doMovement();

		if(enemyBulletFired > 0) {
			g.setColor(Color.WHITE);
			g.drawString(sprintf("Enemy Bullet Intercept Rate: %.4f%%",enemyBulletIntercepted/(double)enemyBulletFired*100), 20, 8+18);
			g.drawString(sprintf("Enemy Bullet Hit Rate: %.4f%%",enemyBulletHit/(double)enemyBulletFired*100), 20, 8+12+18);
		}
	}

	private LinkedList<Bullet> bullets = new LinkedList<Bullet>();

	@Override
	public void onBulletFired(Bullet b) {
		super.onBulletFired(b);

		bullets.add(b);

		//calculate where it will be on all future waves
		calculateShadowsForBullet(b);
	}

	private void calculateShadowsForWave(Wave<Data> wave) {
		if(wave.imaginary) return;

		Iterator<Bullet> it = bullets.iterator();
		while(it.hasNext()) {
			Bullet b = it.next();
			if(!b.isActive()) {
				it.remove();
				continue;
			}
			calculateShadow(wave,b);
		}
	}

	private void calculateShadowsForBullet(Bullet b) {
		for(Wave<Data> wave : enemyWaves) {
			if(wave.imaginary)
				continue;

			calculateShadow(wave,b);
		}
	}

	private void calculateShadow(Wave<Data> wave, Bullet b) {
		//until bullet is past wave calculate ahead
		long timeOffset = 0;

		double x = b.getX();
		double y = b.getY();
		double h = b.getHeadingRadians();
		double v = b.getVelocity();

		do {
			double r = wave.getRadius(time + timeOffset);
			Line line = Line.projection(x, y, h, v*timeOffset, v*(timeOffset+1));

			if(myPosition.distanceSq(line.x1, line.y1) > wave.distanceSq(myPosition) - r*r) {
				break;
			}

			wave.shadowBullet(b, line, time + timeOffset);
		} while(++timeOffset < 110);
	}

	private void removeShadow(Bullet b) {
		for(Wave<Data> wave : enemyWaves) {
			//check if bullet has yet to pass through wave
			double r = wave.getRadius(time);
			double d = wave.distanceSq(myPosition) - r*r;

			if(d > myPosition.distanceSq(b.getX(), b.getY())) {
				//if it hasn't, remove it from the wave
				wave.removeShadow(b);
			}
		}
	}

	private void checkWaves() {
		Iterator<Wave<Data>> waveIterator = enemyWaves.iterator();
		while(waveIterator.hasNext()) {
			Wave<Data> wave = waveIterator.next();

			double radius = wave.getRadius(time);

			/* /////////////////////// */
			/* DRAW WAVE */
			g.setColor(new Color(128,128,128));
			double escape = abs(wave.escapeAngle);
			if(wave.imaginary) {
				g.setColor(new Color(255,0,0));
			}
			g.drawArc(
					(int)(wave.x - radius),
					(int)(wave.y - radius),
					(int)(radius*2),
					(int)(radius*2),
					(int)Math.toDegrees(wave.directAngle-escape),
					(int)Math.toDegrees(escape*2));


			//draw shadows
			g.setColor(Color.GREEN);
			for(double[] range : wave.bulletShadows) {
				g.draw(Line.projection(wave.x, wave.y,
						range[0]*wave.escapeAngle+wave.directAngle,
						radius, radius+wave.speed));

				g.draw(Line.projection(wave.x, wave.y,
						range[1]*wave.escapeAngle+wave.directAngle,
						radius, radius+wave.speed));


			}

			/* END DRAW WAVE */
			/* /////////////////////// */

			if(wave.imaginary) {
				/*
				 * if the wave is imaginary (a heat wave)
				 * we should remove it automatically after
				 * 2 ticks which is when we will detect their
				 * actual wave, if they fire
				 */
				if(!isAlone && time - wave.fireTime > 2) {
					waveIterator.remove();
				}
			} else
				/* Clean up waves that can no longer hit us */
				if(wave.didIntersect(myPosition, time)) {
					onWaveRemoval(wave,false);
					waveIterator.remove();
				}
		}
	}

	private Wave<Data> findBestSurfableWave(Wave<Data> exclude) {
		double halfBotWidth = 18 + Math.sin(lastEnemyPosition.angleTo(myPosition))*7.4558441;
		//find the wave soonest to hit us
		Wave<Data> wave = null;
		double bestETA = Double.POSITIVE_INFINITY;
		for(Wave<Data> check : enemyWaves) {
			if(check == exclude)
				continue;

			double distance = myPosition.distance(check) - check.getRadius(time);
			double ETA = distance / check.speed;

			if(distance < 0 && Math.abs(distance) + check.speed > halfBotWidth)
				continue;

			if(ETA < bestETA) {
				wave = check;
				bestETA = ETA;
			}
		}

		return wave;
	}


	private void checkMove(Simulate sim, Wave<Data> wave, int orbitDirection) {
		Move move = predictMove(sim.position,wave,
				sim.heading,sim.velocity,orbitDirection);

		sim.angleToTurn = move.angleToTurn;
		sim.direction = move.direction;
		sim.maxVelocity = min(sim.maxVelocity, move.maxVelocity);
	}

	/**
	 * Calculate the driving
	 */
	protected Move predictMove(Vector myPosition, Vector orbitCenter,
			double myHeading, double myVelocity, int orbitDirection) {
		Move data = new Move();

		/* Better safe then very very sorry */
		if(orbitDirection == 0)
			orbitDirection = 1;

		double angleToRobot = orbitCenter.angleTo(myPosition);

		/*
		 * if the orbit direction is clockwise/counter, we want to try and
		 * point our robot in that direction, which is why we multiply by it
		 */
		double travelAngle = angleToRobot + (Math.PI/2.0) * orbitDirection;

		/* DONE add distancing to drive method */
		/* TODO add a better distancing method */
		double distance = myPosition.distance(orbitCenter);
		double distancing = ((distance-distBestDistance)/distBestDistance);

		double limit = limit(
				distMinRotation,
				distMultiplier*distance+distAdder,
				distMaxRotation);

		distancing = limit(-limit, distancing, limit);

		travelAngle += distancing*orbitDirection;

		travelAngle = Tools.fastSmooth(myPosition, travelAngle, orbitDirection,
				distance, fieldWidth, fieldHeight);

		data.angleToTurn = Utils.normalRelativeAngle(travelAngle - myHeading);
		data.direction = 1;

		/*
		 * If our backend is closer to direction, use that instead, and
		 * inform the caller that we are going to be going in reverse instead
		 */
		if(abs(data.angleToTurn) > Math.PI/2.0) {
			data.angleToTurn = Utils.normalRelativeAngle(data.angleToTurn - Math.PI);
			data.direction = -1;
		}

		/*
		 * Slow down so we do not ram head long into the walls and can instead turn to avoid them
		 * Still here (just in case)
		 */
		if(!field.contains(myPosition.projectNew(myHeading, myVelocity*3.25)))
			data.maxVelocity = 0;

		if(!field.contains(myPosition.projectNew(myHeading, myVelocity*5)))
			data.maxVelocity = 4;

		return data;
	}

	private void doMovement() {
		Wave<Data> wave = findBestSurfableWave(null);
		if(wave == null || DEBUG_WAVELESS_MOVEMENT) {
			doWavelessMovement();
			return;
		}

		g.setColor(Color.WHITE);
		g.draw(field);

		doSurfingMovement(wave);
	}

	private void doSurfingMovement(Wave<Data> wave) {
		/* Do the actual surfing and such */
		setMaxVelocity(Rules.MAX_VELOCITY);

		Risk[] directions = new Risk[] {
				checkWaveRisk(wave,orbitDirection,0),
				checkWaveRisk(wave,orbitDirection,Rules.MAX_VELOCITY),
				checkWaveRisk(wave,-orbitDirection,Rules.MAX_VELOCITY),
		};

		int bestIndex = 0;
		double minRisk = Double.POSITIVE_INFINITY;
		for(int i = 0; i < directions.length; ++i) {
			if(directions[i].risk < minRisk) {
				bestIndex = i;
				minRisk = directions[i].risk;
			}
		}

		Move move = predictMove(myPosition,wave,
				status.getHeadingRadians(),status.getVelocity(),
				directions[bestIndex].orbitDirection );


		double maxVelocity = min(directions[bestIndex].maxVelocity,move.maxVelocity);
		setMaxVelocity(maxVelocity);
		setTurnBody(move.angleToTurn);
		setMove(1000*move.direction);
		onMoveDecided(maxVelocity,move.angleToTurn,1000*move.direction);
	}

	protected void onMoveDecided(double maxVelocity, double angleToTurn, double toMove) {

	}

	private Risk checkWaveRisk(Wave<Data> wave, int orbitDirection, double maxVelocity) {
		/*
		 * Simplify the output to our direction chooser
		 */
		Risk waveRisk = new Risk();
		waveRisk.orbitDirection = orbitDirection;
		waveRisk.maxVelocity = maxVelocity;

		/*
		 * Create a simulator for our movement
		 */
		Simulate sim = createSimulator();
		sim.maxVelocity = maxVelocity;
		sim.direction = orbitDirection;

		/*
		 * Used for our distance danger
		 */
		double currentDistance = wave.distance(sim.position);
		double predictedDistance = 0;
		int intersected = 0;

		long timeOffset = 0;

		/*
		 * Reset the factors, so we do not get cross risk check contamination
		 */
		wave.resetFactors();

		g.setColor(Color.ORANGE);

		/*
		 * Since our radar is limited to 1200 and the slowest bullet moves at
		 * 11 per second, calculating past 110 is pointlessly impossible, so
		 * it is a safe number to use as our break point.
		 */
		while(timeOffset < 110) {
			//DRAW DANGER MOVEMENT
			g.drawOval((int)sim.position.x-3, (int)sim.position.y-3, 6, 6);

			if(wave.doesIntersect(sim.position, time+timeOffset)) {
				++intersected;
				predictedDistance += wave.distance(sim.position);
			} else if(intersected > 0) {
				predictedDistance /= intersected;

				waveRisk.risk += checkIntersectionRisk(wave);

				Wave<Data> second = findBestSurfableWave(wave);
				if(second != null) {

					//do second wave checking
					Simulate reverse = sim.copy();
					reverse.direction = -orbitDirection;

					double bestRisk = Double.POSITIVE_INFINITY;
					boolean doBreak = false;

					final long timeLimit = timeOffset + 30;

					g.setColor(Color.RED);

					int n0 = 0;
					int n1 = 0;
					while(timeOffset < timeLimit) {
						//for every position along reverse and sim, calculate the PI
						double risk = checkStandingRisk(second,reverse.position);
						if(risk > 0 && risk < bestRisk) {
							bestRisk = risk;
						}

						if(second.intersects(reverse.position, time+timeOffset)) {
							++n0;
						} else if(n0 != 0) {
							doBreak = true;
						}

						risk = checkStandingRisk(second,sim.position);
						if(risk > 0 && risk < bestRisk) {
							bestRisk = risk;
						}

						if(second.intersects(reverse.position, time+timeOffset)) {
							++n1;
						} else if(n1 != 0) {
							if(doBreak) {
								break;
							}
						}

						g.drawRect((int)reverse.position.x-2, (int)reverse.position.y-2, 4, 4);
						g.drawRect((int)sim.position.x-2, (int)sim.position.y-2, 4, 4);

						checkMove(reverse,wave,-orbitDirection);
						reverse.step();
						reverse.maxVelocity = maxVelocity;

						checkMove(sim,wave,orbitDirection);
						sim.step();
						sim.maxVelocity = maxVelocity;

						++timeOffset;
					}

					if(bestRisk < 4) {
						waveRisk.risk += bestRisk/2.0;
					}
				}

				break;
			}

			checkMove(sim,wave,orbitDirection);

			sim.step();
			sim.maxVelocity = maxVelocity;

			++timeOffset;
		}

		double distanceRisk = currentDistance / predictedDistance;
		distanceRisk *= distanceRisk;
		waveRisk.risk *= distanceRisk;

		return waveRisk;
	}

	private double checkStandingRisk(Wave<Data> wave, Vector target) {
		wave.resetFactors();
		wave.standingIntersection(target);
		return checkIntersectionRisk(wave);
	}

	private long cacheTime = -1;
	private Wave<Data> cachedListsWave = null;
	private List<Entry<Data>> cachedList = null;

	private double checkIntersectionRisk(Wave<Data> wave) {
		double risk = 0;
		double factorCenter = (wave.minFactor + wave.maxFactor) / 2.0;

		List<Entry<Data>> list = null;
		if(cacheTime == time && cachedListsWave == wave) {
			list = cachedList;
		} else {
			list = tree.nearestNeighbor(wave.data.getWeightedDataArray(), 64, false);
		}

		int iterate = 0;

		double hitrate = enemyBulletHit/(double)enemyBulletFired;

		//how much do the shadows cover our min to max risk area
		for(Entry<Data> e : list) {
			Data data = e.value;

			if(e.distance > 8+hitrate*24 && iterate != 0)
				continue;

			++iterate;

			double eRisk = 0;

			eRisk = 0.1/(1.0+abs(data.guessfactor-factorCenter));
			if(wave.minFactor < data.guessfactor && wave.maxFactor > data.guessfactor) {
				eRisk += 0.9;
			}

			double factorRange = wave.maxFactor - wave.minFactor;
			double coveredRange = 0;

			for(double[] intr : wave.bulletShadows) {
				if(intr[0] >= wave.maxFactor || intr[1] <= wave.minFactor)
					continue;
				double minf = intr[0];
				double maxf = intr[1];
				if(minf < wave.minFactor) minf = wave.minFactor;
				if(maxf > wave.maxFactor) maxf = wave.maxFactor;
				coveredRange += maxf - minf;
			}

			double ratio = coveredRange / factorRange;

			eRisk *= 1.0 - ratio;
			double timeWeight = AverageSingleRoundTime/(double)(AverageSingleRoundTime + globalTime - data.globalTime);

			double weight = (data.weight / (1.0+e.distance)) * timeWeight;

			risk += eRisk*weight;
		}

		cachedList = list;
		cachedListsWave = wave;
		cacheTime = time;

		return risk / list.size();
	}

	private void doWavelessMovement() {
		final int initialTurns = (int)Math.ceil(3.0/coolingRate)+4;
		double safeTurns = status.getGunHeat()/coolingRate;

		if(time < initialTurns) {
			/* Do we have enough time to move around before they can start firing? */
			if(safeTurns > TimeBeforeFirstWaveToStartTurning || DEBUG_WAVELESS_MOVEMENT) {
				doMinimalRiskMovement();
			} else {
				Move data = predictMove(myPosition,lastEnemyPosition,
						status.getHeadingRadians(),status.getVelocity(),orbitDirection);
				setTurnBody(data.angleToTurn);
				setMaxVelocity(0);

				setMove(0);
				//peer.setTurnBody(0);
				setMaxVelocity(0);

				onMoveDecided(0,data.angleToTurn,0);
			}
		} else {
			/*
			 * Things are no longer safe, set our movement to `sane`
			 * 
			 * TODO: RAM if enemy is disabled
			 */
			if(status.getOthers() == 0) {
				/*
				 * No bullets in the air, enemy dead, we can do our victory DANCE (YAY!)
				 */
				doVictoryDance();
			} else {
				//Some other thing to do when nothing is in the air??
				doMinimalRiskMovement();
			}
		}
	}

	protected void doVictoryDance() {}

	protected void doMinimalRiskMovement() {
		double heading = status.getHeadingRadians();
		double velocity = status.getVelocity();

		Rectangle escapeField = new Rectangle(30,30,fieldWidth-60,fieldHeight-60);
		g.setColor(Color.WHITE);
		g.draw(escapeField);


		if(DEBUG_WAVELESS_MOVEMENT) {
			Wave.hitbox.setFrame(lastEnemyPosition, 36, 36);
			g.setColor(Color.RED);
			g.draw(Wave.hitbox);
		}

		//Do minimal risk movement
		Vector target = myPosition.clone();
		Vector bestTarget = myPosition;
		double angle = 0;
		double bestRisk = checkWavelessRisk(escapeField,bestTarget);
		double enemyDistance = myPosition.distance(lastEnemyPosition);

		//a little dynamic distancing
		//enemyDistance += 18*max((enemyDistance-36-50)/100.0,1.0);
		enemyDistance += limit(-18, -24.48+0.18*enemyDistance ,18);
		//g.draw(new Line(target,target.projectNew(heading, signum(velocity)*200)));

		while(angle < Math.PI*2) {
			double targetDistance = min(200,enemyDistance);

			target.setLocationAndProject(myPosition, angle, targetDistance);
			if(escapeField.contains(target)) {

				// a little nudge to perfer the direction its already going to
				// help avoid both equal opposite jam issues

				double risk = checkWavelessRisk(escapeField,target);

				if(risk < bestRisk) {
					bestRisk = risk;
					bestTarget = target.clone();
				}
				g.setColor(Color.BLUE);
				g.drawRect((int)target.x-2, (int)target.y-2, 4, 4);

			}
			angle += Math.PI/32.0;
		}
		g.setColor(bodyColor);
		g.drawRect((int)bestTarget.x-2, (int)bestTarget.y-2, 4, 4);

		double travelAngle = myPosition.angleTo(bestTarget);

		double forward = myPosition.distance(bestTarget);

		double angleToTurn = Utils.normalRelativeAngle(travelAngle - status.getHeadingRadians());
		int direction = 1;

		if(abs(angleToTurn) > Math.PI/2.0) {
			angleToTurn = Utils.normalRelativeAngle(angleToTurn - Math.PI);
			direction = -1;
		}

		//Slow down so we do not ram head long into the walls and can instead turn to avoid them
		double maxVelocity = Rules.MAX_VELOCITY;

		if(!field.contains(myPosition.projectNew(heading, velocity*3.25)))
			maxVelocity = 0;

		if(!field.contains(myPosition.projectNew(heading, velocity*5)))
			maxVelocity = 4;

		if(angleToTurn > 0.7 && velocity < 7) {
			maxVelocity = 0;
		}

		if(DEBUG_WAVELESS_MOVEMENT) {
			g.setColor(Color.WHITE);
			g.drawString("v "+status.getVelocity(), (float)myPosition.x-16, (float)myPosition.y-36-12);
		}

		setMaxVelocity(maxVelocity);
		setTurnBody(angleToTurn);
		setMove(forward*direction);

		onMoveDecided(maxVelocity,angleToTurn,forward*direction);
	}

	private double checkWavelessRisk(Rectangle field, Vector pos) {
		double risk = 100.0/pos.distanceSq(lastEnemyPosition);

		for(double[] edge : field.getEdges()) {
			risk += 5.0/(1.0+Line2D.ptSegDistSq(edge[0], edge[1], edge[2], edge[3], pos.x, pos.y));
		}

		/*
		 * get points between enemy location and corner and add risk!!!!
		 * these are bad places to be! Our hitbox is larger here if nothing else!
		 */
		for(double[] corner : field.getCorners()) {
			corner[0] = (corner[0]+lastEnemyPosition.x)/2.0;
			corner[1] = (corner[1]+lastEnemyPosition.y)/2.0;
			if(lastEnemyPosition.distanceSq(corner[0], corner[1]) < 22500 /* 150 */) {
				g.drawRect((int)corner[0]-2, (int)corner[1]-2, 4, 4);
				risk += 5.0/(1.0+pos.distanceSq(corner[0], corner[1]));
			}
		}

		return risk;
	}

	/**
	 * Account for our bullets hitting the enemy
	 */
	@Override
	public void onBulletHit(BulletHitEvent e) {
		super.onBulletHit(e);

		/* Get the power of the bullet of ours that hit */
		double bulletPower = e.getBullet().getPower();

		/* Determine how much damage our bullet does to the enemy
		 * and adjust our stored copy to reflect the amount lost
		 */
		lastEnemyEnergy -= Rules.getBulletDamage(bulletPower);
	}

	/**
	 * Account for one of our bullets hitting an enemy bullet.
	 */

	@Override
	public void onBulletHitBullet(BulletHitBulletEvent e) {
		super.onBulletHitBullet(e);

		++enemyBulletIntercepted;
		removeShadow(e.getBullet());
		handleBullet(e.getHitBullet(),true);
	}

	/**
	 * Account for one of the enemies bullets hitting us.
	 */
	@Override
	public void onHitByBullet(HitByBulletEvent e) {
		super.onHitByBullet(e);

		/*
		 * Increase the enemy energy based on how powerful the bullet that
		 * hit us was this is so we can get reliable energy drop detection
		 */
		lastEnemyEnergy += Rules.getBulletHitBonus(e.getPower());

		++enemyBulletHit;
		handleBullet(e.getBullet(),false);
	}

	/**
	 * Handle an actual enemy bullet
	 */
	private void handleBullet(Bullet b, boolean collision) {
		Vector bulletPosition = new Vector(b.getX(),b.getY());

		if(DEBUG_BULLET_HANDLING) {
			if(collision) {
				println(time + "- Bullet Bullet Collision");
			} else {
				println(time + "- Bullet Robot Collision");
			}
		}

		double bulletDistToEnemy = lastEnemyPosition.distance(bulletPosition);

		/* Find the matching wave */
		Iterator<Wave<Data>> wi = enemyWaves.iterator();
		while(wi.hasNext()) {
			Wave<Data> wave = wi.next();

			if(wave.imaginary && time - wave.fireTime < 3) {
				if(!collision || bulletDistToEnemy > 2)
					continue;
			}
			//if(wave.imaginary && time - wave.fireTime < 3 && distToEnemy > wave.speed) {
			//	continue;
			//}

			/* check if the power is close enough to be our wave. This
			 * margin is small, only to allow room for rounding errors
			 */
			if(abs(b.getPower()-wave.power) < 0.001 ||
					(wave.imaginary && abs(b.getPower()-wave.power) < 0.5)) {
				double d = wave.distanceSq(bulletPosition);

				/* the current distance of the wave */
				boolean match = false;

				if(collision) {
					double rad1 = wave.speed*((time-1) - wave.fireTime);
					double rad2 = wave.speed*((time-2) - wave.fireTime);

					if(abs(rad1*rad1-d) < 128) {
						if(DEBUG_BULLET_HANDLING) println("Match on BD1");
						match = true;
					} else if(abs(rad2*rad2-d) < 128) {
						if(DEBUG_BULLET_HANDLING) println("Match on BD2");
						match = true;
					}

				} else {
					double rad0 = wave.speed*(time - wave.fireTime);
					if(abs(d-rad0*rad0) < 128) {
						match = true;
					}
				}

				if(match) {
					/* Alternate method, but why recalculate something? Could be used as an extra check. */
					/* updateTree(wave,wave.angleTo(bulletPosition)); */
					updateTree(wave,b.getHeadingRadians());

					onWaveRemoval(wave,true);

					wi.remove();

					if(wave.imaginary) //just to keep everything nice and balanced
						++enemyBulletFired;

					return;
				}
			}
		}

		/* If not found say so, cause something f*cked up! */
		//println("Error: Unknown Bullet Collision");
		//println("\tBullet:["+b.getX()+","+b.getY()+"] @ h"+b.getHeadingRadians()+" @ p" + b.getPower());

		printf("Error: Unknown Bullet Collision\n\tBullet:[%.4f,%.4f] @ h%.4f @ p%.4f\n\tEnemy:[%.4f,%.4f] D To B %.4f\nWaves:\n",
				b.getX(),b.getY(),b.getHeadingRadians(),b.getPower(),lastEnemyPosition.x,lastEnemyPosition.y,bulletDistToEnemy);

		for(Wave<Data> wave : enemyWaves) {
			double rad = wave.speed*(time - wave.fireTime);
			double rad1 = wave.speed*((time-1) - wave.fireTime);
			double d = wave.distanceSq(bulletPosition);
			printf("\tp %.4f  d %.4f  d1 %.4f  i %b\n",wave.power,abs(rad*rad-d),abs(rad1*rad1-d),wave.imaginary);
		}

		//enemyEnergyHistory
	}

	protected void onWaveRemoval(Wave<Data> wave, boolean hit) {}

	private void updateTree(Wave<Data> wave, double angleToBullet) {
		double angleOffset =
			Utils.normalRelativeAngle(angleToBullet - wave.directAngle);

		wave.data.guessfactor = angleOffset / wave.escapeAngle;

		tree.addPoint(wave.data.getWeightedDataArray(), wave.data);
	}
}

/**
 * Recorded State
 */
class State {
	Vector myPosition;
	Vector enemyPosition;
	Data data = new Data();
}

/**
 * Data gained from movement prediction
 */
class Move {
	double angleToTurn = 0;
	double maxVelocity = Rules.MAX_VELOCITY;
	int direction = 1;
}

/**
 * Risk of direction
 */
class Risk {
	double risk;
	double maxVelocity;
	int orbitDirection;
}
