package voidious.gun;

import java.awt.Color;
import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Vector;

import ags.utils.KdTree.Entry;

import robocode.util.Utils;

import voidious.gfx.ColoredValueSet;
import voidious.gfx.RoboGraphic;
import voidious.utils.BattleField;
import voidious.utils.DiaUtils;
import voidious.utils.DiaWave;
import voidious.utils.KnnView;
import voidious.utils.DiaWave.FiringTarget;
import voidious.utils.DistanceFormula;
import voidious.utils.RobotState;
import voidious.utils.TimestampedFiringAngleSet;
import voidious.utils.geom.LineSeg;

/**
 * Copyright (c) 2011 - Voidious
 *
 * This software is provided 'as-is', without any express or implied
 * warranty. In no event will the authors be held liable for any damages
 * arising from the use of this software.
 *
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, and to alter it and redistribute it
 * freely, subject to the following restrictions:
 *
 *    1. The origin of this software must not be misrepresented; you must not
 *    claim that you wrote the original software.
 *
 *    2. Altered source versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 *
 *    3. This notice may not be removed or altered from any source
 *    distribution.
 */

public class AntiSurferSimGun implements DuelGun<TimestampedFiringAngleSet> {
    private static final int MIN_SIM_TICKS = 3;
    private static final int MAX_SIM_TICKS = 13;
    private static final int FIRING_ANGLES = 59;

    private static final String VIEW_NAME_HITS = "simsurf_hits";

    private Map<String, EnemyGunData> _enemies;
    private BattleField _battleField;
    private Vector<RoboGraphic> _renderables;
    private Map<DiaWave, Double> _firingAngles;
    private int _enemiesTotal;
    private List<String> _visitViewNames;
    private DistanceFormula _formula;
    private List<DiaWave> _waves;

    public AntiSurferSimGun(Map<String, EnemyGunData> enemies, 
            BattleField battleField, Vector<RoboGraphic> renderables, 
            int enemiesTotal, List<DiaWave> waves) {
        _enemies = enemies;
        _battleField = battleField;
        _renderables = renderables;
        _firingAngles = new HashMap<DiaWave, Double>();
        _enemiesTotal = enemiesTotal;
        _visitViewNames = new ArrayList<String>();
        _formula = new AntiSurferFormula(enemiesTotal);
        _waves = waves;
    }

    @Override
    public void clear() {
        _firingAngles.clear();
    }

    @Override
    public void clearWave(DiaWave w) {
        _firingAngles.remove(w);
    }

    @Override
    public String getLabel() { return "Anti-Surfer Sim Gun"; }

    @Override
    public double aimWithWave(DiaWave w, boolean painting) {
        if (_firingAngles.containsKey(w)) {
            return _firingAngles.get(w);
        }

        EnemyGunData enemyData = _enemies.get(w.botName);
        double baseMinAngle = -w.preciseEscapeAngle(
            w.orbitDirection == DiaWave.COUNTERCLOCKWISE);
        double baseMaxAngle = w.preciseEscapeAngle(
            w.orbitDirection == DiaWave.CLOCKWISE);
        double minAngle = baseMinAngle;
        double maxAngle = baseMaxAngle;

        DiaWave surfWave = getSimSurfWave(w);
        KnnView<TimestampedFiringAngleSet> hitsView =
            enemyData.views.get(VIEW_NAME_HITS);
        if (surfWave != null && hitsView.size() > 0) {
            RobotState startState = new RobotState(w.targetLocation,
                w.targetHeading, w.targetVelocity, w.fireTime);
            FiringTarget surfTargetClock = surfWave.calculatePreciseEscapeAngle(
                surfWave.orbitDirection == DiaWave.CLOCKWISE, startState);
            FiringTarget surfTargetCounter = surfWave.calculatePreciseEscapeAngle(
                surfWave.orbitDirection == DiaWave.COUNTERCLOCKWISE,
                startState);
            double surfMaxAngle = Utils.normalRelativeAngle(
                DiaUtils.absoluteBearing(surfWave.sourceLocation,
                surfTargetClock.location) - surfWave.absBearing);
            double surfMinAngle = Utils.normalRelativeAngle(
                DiaUtils.absoluteBearing(surfWave.sourceLocation,
                surfTargetCounter.location) - surfWave.absBearing);
            if (painting) {
                _renderables.add(RoboGraphic.drawPoint(surfTargetClock.location,
                    Color.magenta));
                _renderables.add(RoboGraphic.drawPoint(
                    surfTargetCounter.location, Color.magenta));
            }

            int kSize = DiaUtils.limit(1, hitsView.size() / hitsView.kDivisor,
                hitsView.kSize);
            List<Entry<TimestampedFiringAngleSet>> surfNeighbors =
                hitsView.nearestNeighbors(surfWave, false, kSize);

            int numScans = surfNeighbors.size();
            double[] hitAngles = new double[numScans];
            double[] hitWeights = new double[numScans];
            Arrays.fill(hitWeights, 1);

            for (int x = 0; x < numScans; x++) {
                if (_enemiesTotal == 1) {
                    double guessFactor = surfNeighbors.get(x).value.guessFactor;
                    hitAngles[x] = Utils.normalRelativeAngle(
                        (guessFactor * surfWave.orbitDirection *
                            surfWave.preciseEscapeAngle(guessFactor >= 0)));
                } else {
                    Point2D.Double dispVector = surfNeighbors.get(x).value
                        .displacementVector;
                    Point2D.Double projectedLocation =
                        surfWave.projectLocation(dispVector);
                    if (!_battleField.rectangle.contains(projectedLocation)) {
                        hitAngles[x] = DiamondFist.NO_FIRING_ANGLE;
                    } else {
                        hitAngles[x] = Utils.normalRelativeAngle(
                            surfWave.firingAngleFromTargetLocation(projectedLocation) 
                                - surfWave.absBearing);
                    }
                }
                hitWeights[x] = 1 / Math.sqrt(surfNeighbors.get(x).distance);
            }

            hitsView.setDecayWeights(surfNeighbors, hitWeights);

            double[] realAngles = DiaUtils.generateFiringAngles(FIRING_ANGLES,
                surfWave.maxEscapeAngle);

            double bandwidth =
                DiaUtils.botWidthAimAngle(surfWave.targetDistance);
            double safestDensity = Double.POSITIVE_INFINITY;
            double safestAngle = (surfMinAngle + surfMaxAngle) / 2;
            for (int x = 0; x < FIRING_ANGLES; x++) {
                if (realAngles[x] < surfMinAngle ||
                        realAngles[x] > surfMaxAngle) {
                    continue;
                }
                double density = 0;
                for (int y = 0; y < numScans; y++) {
                    if (hitAngles[y] == DiamondFist.NO_FIRING_ANGLE) {
                        continue;
                    }

                    double ux = (realAngles[x] - hitAngles[y]) / bandwidth;
                    density += Math.exp(-0.5 * ux * ux) * hitWeights[y];
                }

                if (density < safestDensity) {
                    safestAngle = realAngles[x];
                    safestDensity = density;
                }
            }

            Point2D.Double safeLineEndpoint = DiaUtils.project(
                surfWave.sourceLocation,
                surfWave.absBearing + safestAngle, 2000);
            LineSeg safestLine = new LineSeg(surfWave.sourceLocation,
                safeLineEndpoint);
            Point2D.Double safeDirectionEndpoint =
                Utils.normalRelativeAngle(surfWave.absBearing + safestAngle
                    - DiaUtils.absoluteBearing(surfWave.sourceLocation,
                        w.targetLocation))
                    > 0 ? 
                surfTargetClock.location : surfTargetCounter.location;
            LineSeg movementLine = new LineSeg(w.targetLocation,
                safeDirectionEndpoint);
            Point2D.Double safePoint = safestLine.intersects(movementLine);
            if (painting) {
                _renderables.add(RoboGraphic.drawLine(surfWave.sourceLocation,
                    safeLineEndpoint, Color.yellow));
                _renderables.add(RoboGraphic.drawText(
                    Double.toString(DiaUtils.round(safestAngle, 3)),
                    surfWave.sourceLocation.x + 30,
                    surfWave.sourceLocation.y - 20, Color.white));
                _renderables.add(RoboGraphic.drawText(
                    Double.toString(DiaUtils.round(surfMinAngle, 3)),
                    surfWave.sourceLocation.x,
                    surfWave.sourceLocation.y - 40, Color.magenta));
                _renderables.add(RoboGraphic.drawText(
                    Double.toString(DiaUtils.round(surfMaxAngle, 3)),
                    surfWave.sourceLocation.x + 60,
                    surfWave.sourceLocation.y - 40, Color.magenta));
                _renderables.add(RoboGraphic.drawLine(w.targetLocation,
                    safeDirectionEndpoint, Color.blue));
                _renderables.add(RoboGraphic.drawPoint(safePoint,
                    Color.green));
            }

            if (safePoint != null) {
                RobotState safeState = new RobotState(safePoint,
                    w.targetHeading, w.targetVelocity, safestAngle > 0 ?
                        surfTargetClock.time : surfTargetCounter.time);
                FiringTarget safeTargetCounter = w.calculatePreciseEscapeAngle(
                    w.orbitDirection == DiaWave.COUNTERCLOCKWISE, safeState);
                FiringTarget safeTargetClock = w.calculatePreciseEscapeAngle(
                    w.orbitDirection == DiaWave.CLOCKWISE, safeState);
                if (painting) {
                    _renderables.add(RoboGraphic.drawLine(safePoint,
                        safeTargetCounter.location, Color.red));
                    _renderables.add(RoboGraphic.drawPoint(
                        safeTargetCounter.location, Color.red));
                    _renderables.add(RoboGraphic.drawLine(safePoint,
                        safeTargetClock.location, Color.red));
                    _renderables.add(RoboGraphic.drawPoint(
                        safeTargetClock.location, Color.red));
                }
                double simMinAngle = Utils.normalRelativeAngle(
                    DiaUtils.absoluteBearing(w.sourceLocation,
                    safeTargetCounter.location) - w.absBearing);
                double simMaxAngle = Utils.normalRelativeAngle(
                    DiaUtils.absoluteBearing(w.sourceLocation,
                    safeTargetClock.location) - w.absBearing);
                minAngle = Math.max(minAngle, simMinAngle);
                maxAngle = Math.min(maxAngle, simMaxAngle);
            }
        }

        List<Entry<TimestampedFiringAngleSet>> nearestNeighbors = null;
        double[] neighborWeights = null;
        for (String viewName : _visitViewNames) {
            KnnView<TimestampedFiringAngleSet> view =
                enemyData.views.get(viewName);
            if (view.size() < view.kDivisor) {
                continue;
            }

            int kSize = DiaUtils.limit(1, view.size() / view.kDivisor,
                view.kSize);
            List<Entry<TimestampedFiringAngleSet>> thisNeighbors =
                view.nearestNeighbors(w, DiamondFist.AIMING, kSize);
            double[] thisWeights = new double[thisNeighbors.size()];
            Arrays.fill(thisWeights, view.weight);

            view.setDecayWeights(thisNeighbors, thisWeights);

            if (nearestNeighbors == null) {
                nearestNeighbors = thisNeighbors;
                neighborWeights = thisWeights;
            } else {
                int totalNeighbors = 
                    thisNeighbors.size() + nearestNeighbors.size();
                for (int x = 0; x < thisNeighbors.size(); x++) {
                    nearestNeighbors.add(thisNeighbors.get(x));
                }
                double[] combinedWeights = new double[totalNeighbors];
                for (int x = 0; x < neighborWeights.length; x++) {
                    combinedWeights[x] = neighborWeights[x];
                }
                for (int x = 0; x < thisNeighbors.size(); x++) {
                    combinedWeights[x + neighborWeights.length] = 
                        thisWeights[x];
                }
                neighborWeights = combinedWeights;
            }
        }

        if (nearestNeighbors == null || nearestNeighbors.size() == 0) {
            return w.absBearing;
        }

        int numScans = nearestNeighbors.size();
        double[] firingAngles = new double[numScans];
        for (int x = 0; x < numScans; x++) {
            if (_enemiesTotal == 1) {
                double guessFactor = nearestNeighbors.get(x).value.guessFactor;
                firingAngles[x] = Utils.normalRelativeAngle(
                    (guessFactor * w.orbitDirection *
                        w.preciseEscapeAngle(guessFactor >= 0)));
            } else {
                Point2D.Double dispVector = nearestNeighbors.get(x).value
                    .displacementVector;
                Point2D.Double projectedLocation =
                    w.projectLocation(dispVector);
                if (!_battleField.rectangle.contains(projectedLocation)) {
                    firingAngles[x] = DiamondFist.NO_FIRING_ANGLE;
                } else {
                    firingAngles[x] = Utils.normalRelativeAngle(
                        w.firingAngleFromTargetLocation(projectedLocation) 
                            - w.absBearing);
                }
            }
        }

        double bestAngle = DiamondFist.NO_FIRING_ANGLE;
        double bestDensity = Double.NEGATIVE_INFINITY;
        double bandwidth = DiaUtils.botWidthAimAngle(
            w.sourceLocation.distance(w.targetLocation)) * 2;
        ColoredValueSet cvs = new ColoredValueSet();
        double[] realAngles =
            DiaUtils.generateFiringAngles(FIRING_ANGLES, w.maxEscapeAngle);
        for (int x = 0; x < FIRING_ANGLES; x++) {
            if (realAngles[x] < minAngle || realAngles[x] > maxAngle) {
                continue;
            }

            double xFiringAngle = realAngles[x];
            double xDensity = 0;
            for (int y = 0; y < numScans; y++) {
                if (firingAngles[y] == DiamondFist.NO_FIRING_ANGLE) {
                    continue;
                }
                double yFiringAngle = firingAngles[y];
                double ux = (xFiringAngle - yFiringAngle) / bandwidth;
                xDensity += Math.exp(-0.5 * ux * ux) * neighborWeights[y];
            }

            if (xDensity > bestDensity) {
                bestAngle = xFiringAngle;
                bestDensity = xDensity;
            }

            if (painting) {
                cvs.addValue(xDensity, w.absBearing + xFiringAngle);
            }
        }

        if (bestAngle == DiamondFist.NO_FIRING_ANGLE) {
            return w.absBearing;
        }

        if (painting) {
            DiamondFist.drawGunAngles(_renderables, w, cvs, 
                w.absBearing + bestAngle, bandwidth);
        }

        double firingAngle = Utils.normalAbsoluteAngle(w.absBearing + bestAngle);
        _firingAngles.put(w, firingAngle);
        return firingAngle;
    }

    @Override
    public List<KnnView<TimestampedFiringAngleSet>> newDataViews() {
        KnnView<TimestampedFiringAngleSet> visitsView =
            new KnnView<TimestampedFiringAngleSet>(_formula)
                .setK(3)
                .setMaxDataPoints(125)
                .setKDivisor(10)
                .visitsOn()
                .virtualWavesOn();
        KnnView<TimestampedFiringAngleSet> visitsView2 =
            new KnnView<TimestampedFiringAngleSet>(_formula)
                .setK(3)
                .setMaxDataPoints(500)
                .setKDivisor(10)
                .visitsOn()
                .virtualWavesOn();
        KnnView<TimestampedFiringAngleSet> visitsView3 =
            new KnnView<TimestampedFiringAngleSet>(_formula)
                .setK(3)
                .setMaxDataPoints(1500)
                .setKDivisor(10)
                .visitsOn()
                .virtualWavesOn();
        KnnView<TimestampedFiringAngleSet> visitsView4 =
            new KnnView<TimestampedFiringAngleSet>(_formula)
                .setK(3)
                .setMaxDataPoints(4000)
                .setKDivisor(10)
                .visitsOn()
                .virtualWavesOn();
        KnnView<TimestampedFiringAngleSet> hitsView =
            new KnnView<TimestampedFiringAngleSet>(_formula)
                .setK(20)
                .setKDivisor(5)
                .setDecayRate(1.6)
                .bulletHitsOn()
                .setName(VIEW_NAME_HITS);
        if (_visitViewNames.isEmpty()) {
            _visitViewNames.add(visitsView.name);
            _visitViewNames.add(visitsView2.name);
            _visitViewNames.add(visitsView3.name);
            _visitViewNames.add(visitsView4.name);
        }
        List<KnnView<TimestampedFiringAngleSet>> views =
            new ArrayList<KnnView<TimestampedFiringAngleSet>>();
        views.add(visitsView);
        views.add(visitsView2);
        views.add(visitsView3);
        views.add(visitsView4);
        views.add(hitsView);

        return views;
    }

    private DiaWave getSimSurfWave(DiaWave aimWave) {
        DiaWave surfWave = null;
        DiaWave nextWave = null;
        int surfIndex = DiaWave.FIRST_WAVE;
        do {
            nextWave = DiaWave.findSurfableWave(_waves, surfIndex++,
                aimWave.targetLocation, aimWave.fireTime);
            if (nextWave != null) {
                long timeToImpact = Math.round(Math.ceil(
                    (nextWave.sourceLocation.distance(aimWave.targetLocation)
                    - nextWave.distanceTraveled(aimWave.fireTime) 
                    - nextWave.bulletSpeed)
                        / nextWave.bulletSpeed));
                if (timeToImpact >= MIN_SIM_TICKS) {
                    if (timeToImpact <= MAX_SIM_TICKS) {
                        surfWave = nextWave;
                    }
                    break;
                }
            }
        } while (nextWave != null && nextWave != aimWave);

        return surfWave;            
    }
}
