package voidious.gun;

import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Vector;

import ags.utils.KdTree.Entry;

import robocode.util.Utils;

import voidious.gfx.ColoredValueSet;
import voidious.gfx.RoboGraphic;
import voidious.utils.BattleField;
import voidious.utils.DiaUtils;
import voidious.utils.DiaWave;
import voidious.utils.KnnView;
import voidious.utils.TimestampedFiringAngleSet;

/**
 * Copyright (c) 2009-2011 - Voidious
 *
 * This software is provided 'as-is', without any express or implied
 * warranty. In no event will the authors be held liable for any damages
 * arising from the use of this software.
 *
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, and to alter it and redistribute it
 * freely, subject to the following restrictions:
 *
 *    1. The origin of this software must not be misrepresented; you must not
 *    claim that you wrote the original software.
 *
 *    2. Altered source versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 *
 *    3. This notice may not be removed or altered from any source
 *    distribution.
 */

public class MainGun implements DuelGun<TimestampedFiringAngleSet> {
    private static final String VIEW_NAME = "Main";
    private static final int K_SIZE = 30;
    private static final int K_DIVISOR = 10;

    private Map<String, EnemyGunData> _enemies;
    private BattleField _battleField;
    private Vector<RoboGraphic> _renderables;
    private Map<DiaWave, Double> _firingAngles;
    private int _enemiesTotal;

    public MainGun(Map<String, EnemyGunData> enemies, BattleField battleField,
            Vector<RoboGraphic> renderables, int enemiesTotal) {
        _enemies = enemies;
        _battleField = battleField;
        _renderables = renderables;
        _firingAngles = new HashMap<DiaWave, Double>();
        _enemiesTotal = enemiesTotal;
    }

    @Override
    public void clear() {
        _firingAngles.clear();
    }

    @Override
    public void clearWave(DiaWave w) {
        _firingAngles.remove(w);
    }

    @Override
    public String getLabel() { return "Main Gun"; }

    @Override
    public double aimWithWave(DiaWave w, boolean painting) {
        if (_firingAngles.containsKey(w)) {
            return _firingAngles.get(w);
        }

        EnemyGunData enemyData = _enemies.get(w.botName);
        KnnView<TimestampedFiringAngleSet> view = enemyData.views.get(VIEW_NAME);

        if (view.size() < 10) {
            return w.absBearing;
        }

        int clusterSize = (int)Math.min(view.size() / K_DIVISOR, view.kSize);
        List<Entry<TimestampedFiringAngleSet>> nearestNeighbors =
            view.nearestNeighbors(w, DiamondFist.AIMING, clusterSize);
        if (nearestNeighbors == null || nearestNeighbors.size() == 0) {
            return w.absBearing;
        }

        int numScans = nearestNeighbors.size();
        double[] firingAngles = new double[numScans];
        for (int x = 0; x < numScans; x++) {
            Point2D.Double dispVector = nearestNeighbors.get(x)
                .value.displacementVector;
            Point2D.Double projectedLocation = w.projectLocation(dispVector);
            if (!_battleField.rectangle.contains(projectedLocation)) {
                firingAngles[x] = DiamondFist.NO_FIRING_ANGLE;
            } else {
                firingAngles[x] = Utils.normalRelativeAngle(
                    w.firingAngleFromTargetLocation(projectedLocation) 
                        - w.absBearing);
            }
        }

        double bestAngle = DiamondFist.NO_FIRING_ANGLE;
        double bestDensity = Double.NEGATIVE_INFINITY;
        double bandwidth = DiaUtils.botWidthAimAngle(
            w.sourceLocation.distance(w.targetLocation)) * 2;
        ColoredValueSet cvs = new ColoredValueSet();

        for (int x = 0; x < numScans; x++) {
            if (firingAngles[x] == DiamondFist.NO_FIRING_ANGLE) { continue; }

            double xFiringAngle = firingAngles[x];

            double xDensity = 0;
            for (int y = 0; y < numScans; y++) {
                if (x == y || firingAngles[y] == DiamondFist.NO_FIRING_ANGLE)
                    continue;

                double yFiringAngle = firingAngles[y];

                double ux = (xFiringAngle - yFiringAngle) / bandwidth;

                // Gaussian
                xDensity += Math.exp(-0.5 * ux * ux);
            }

            if (xDensity > bestDensity) {
                bestAngle = xFiringAngle;
                bestDensity = xDensity;
            }

            if (painting) {
                cvs.addValue(xDensity, w.absBearing + xFiringAngle);
            }
        }

        if (bestAngle == DiamondFist.NO_FIRING_ANGLE) {
            return w.absBearing;
        }

        if (painting) {
            DiamondFist.drawGunAngles(_renderables, w, cvs, 
                w.absBearing + bestAngle, bandwidth);
        }

        double firingAngle =
            Utils.normalAbsoluteAngle(w.absBearing + bestAngle);
        _firingAngles.put(w, firingAngle);

        return firingAngle;
    }

    @Override
    public List<KnnView<TimestampedFiringAngleSet>> newDataViews() {
        List<KnnView<TimestampedFiringAngleSet>> views =
            new ArrayList<KnnView<TimestampedFiringAngleSet>>();
        views.add(new KnnView<TimestampedFiringAngleSet>(
            new MainGunFormula(_enemiesTotal))
                .setK(K_SIZE)
                .visitsOn()
                .virtualWavesOn()
                .meleeOn()
                .setName(VIEW_NAME));
        return views;
    }
}
