package voidious.gun;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Vector;

import ags.utils.KdTree.Entry;

import robocode.util.Utils;

import voidious.gfx.ColoredValueSet;
import voidious.gfx.RoboGraphic;
import voidious.utils.KnnView;
import voidious.utils.DiaUtils;
import voidious.utils.DiaWave;
import voidious.utils.TimestampedFiringAngleSet;

/**
 * Copyright (c) 2009-2011 - Voidious
 *
 * This software is provided 'as-is', without any express or implied
 * warranty. In no event will the authors be held liable for any damages
 * arising from the use of this software.
 *
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, and to alter it and redistribute it
 * freely, subject to the following restrictions:
 *
 *    1. The origin of this software must not be misrepresented; you must not
 *    claim that you wrote the original software.
 *
 *    2. Altered source versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 *
 *    3. This notice may not be removed or altered from any source
 *    distribution.
 */

public class TripHammerKnnGun implements DuelGun<TimestampedFiringAngleSet> {
    private static final String VIEW_NAME = "TripHammerKNN";
    private static final int FIRING_ANGLES = 59;
    private static final int MAX_K_SIZE = 340;
    private static final int K_DIVISOR = 14;

    private Map<String, EnemyGunData> _enemies;
    private Vector<RoboGraphic> _renderables;
    private HashMap<DiaWave, Double> _firingAngles;

    public TripHammerKnnGun(Map<String, EnemyGunData> enemies, 
            Vector<RoboGraphic> renderables) { 
        _enemies = enemies;
        _renderables = renderables;
        _firingAngles = new HashMap<DiaWave, Double>();
    }

    @Override
    public double aimWithWave(DiaWave w, boolean painting) {
        if (_firingAngles.containsKey(w)) {
            return _firingAngles.get(w);
        }

        EnemyGunData enemyData = _enemies.get(w.botName);
        KnnView<TimestampedFiringAngleSet> view =
            enemyData.views.get(VIEW_NAME);

        int desiredKSize = DiaUtils.limit(1, view.size() / K_DIVISOR, 
            MAX_K_SIZE);
        List<Entry<TimestampedFiringAngleSet>> nearestNeighbors =
            view.nearestNeighbors(w, DiamondFist.AIMING, desiredKSize);

        if (enemyData == null || view.size() == 0) {
            return w.absBearing;
        }

        int numScans = nearestNeighbors.size();
        double[] firingAngles = new double[numScans];
        double[] weights = new double[numScans];
        for (int x = 0; x < numScans; x++) {
            double guessFactor = nearestNeighbors.get(x).value.guessFactor;
            firingAngles[x] = Utils.normalRelativeAngle(
                (guessFactor * w.orbitDirection
                * w.preciseEscapeAngle(guessFactor >= 0)));
            weights[x] = 1 / Math.sqrt(nearestNeighbors.get(x).distance);
        }

        double bandwidth = 2 * DiaUtils.botWidthAimAngle(
            w.sourceLocation.distance(w.targetLocation));
        double bestAngle = DiamondFist.NO_FIRING_ANGLE;
        double bestDensity = Double.NEGATIVE_INFINITY;
        ColoredValueSet cvs = new ColoredValueSet();
        if (numScans > FIRING_ANGLES) {
            double[] realAngles =
                DiaUtils.generateFiringAngles(FIRING_ANGLES, w.maxEscapeAngle);

            for (int x = 0; x < FIRING_ANGLES; x++) {
                double density = 0;
                for (int y = 0; y < numScans; y++) {
                    if (firingAngles[y] == DiamondFist.NO_FIRING_ANGLE) {
                        continue;
                    }
                    double ux = (realAngles[x] - firingAngles[y]) / bandwidth;
                    if (Math.abs(ux) < 1) {
                        density +=
                            DiaUtils.square(1 - DiaUtils.square(Math.abs(ux)))
                            * weights[y];
                    }
                }

                if (density > bestDensity) {
                    bestAngle = realAngles[x];
                    bestDensity = density;
                }

                if (painting) {
                    cvs.addValue(density, w.absBearing + realAngles[x]);
                }
            }
        } else {
            for (int x = 0; x < numScans; x++) {
                if (firingAngles[x] == DiamondFist.NO_FIRING_ANGLE) {
                    continue;
                }
                double xFiringAngle = firingAngles[x];
                double xDensity = 0;
                for (int y = 0; y < numScans; y++) {
                    if (x == y ||
                            firingAngles[y] == DiamondFist.NO_FIRING_ANGLE) {
                        continue;
                    }
                    double yFiringAngle = firingAngles[y];
                    double ux = (xFiringAngle - yFiringAngle) / bandwidth;
                    xDensity += Math.exp(-0.5 * ux * ux) * weights[y];
                }

                if (xDensity > bestDensity) {
                    bestAngle = xFiringAngle;
                    bestDensity = xDensity;
                }

                if (painting) {
                    cvs.addValue(xDensity, w.absBearing + xFiringAngle);
                }
            }            
        }

        if (bestAngle == DiamondFist.NO_FIRING_ANGLE) {
            return w.absBearing;
        }

        if (painting) {
            DiamondFist.drawGunAngles(_renderables, w, cvs, 
                w.absBearing + bestAngle, bandwidth);
        }

        double firingAngle =
            Utils.normalAbsoluteAngle(w.absBearing + bestAngle);
        _firingAngles.put(w, firingAngle);

        return firingAngle;
    }

    @Override
    public void clear() { }

    @Override
    public void clearWave(DiaWave w) {
        _firingAngles.remove(w);
    }

    @Override
    public String getLabel() {
        return "Main Gun";
    }

    @Override
    public List<KnnView<TimestampedFiringAngleSet>> newDataViews() {
        List<KnnView<TimestampedFiringAngleSet>> views =
            new ArrayList<KnnView<TimestampedFiringAngleSet>>();
        TripHammerFormula distanceTripHammer = new TripHammerFormula();
        KnnView<TimestampedFiringAngleSet> view =
            new KnnView<TimestampedFiringAngleSet>(distanceTripHammer)
                .visitsOn()
                .virtualWavesOn()
                .setName(VIEW_NAME);
        views.add(view);
        return views;
    }
}
