package voidious.gun;

import robocode.util.Utils;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import voidious.utils.DiaWave;
import voidious.utils.DiaUtils;

/**
 * Copyright (c) 2009-2011 - Voidious
 *
 * This software is provided 'as-is', without any express or implied
 * warranty. In no event will the authors be held liable for any damages
 * arising from the use of this software.
 *
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, and to alter it and redistribute it
 * freely, subject to the following restrictions:
 *
 *    1. The origin of this software must not be misrepresented; you must not
 *    claim that you wrote the original software.
 *
 *    2. Altered source versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 *
 *    3. This notice may not be removed or altered from any source
 *    distribution.
 */

public class VirtualGunsManager<T> {
    private static final double TYPICAL_DISTANCE = 400;
    private static final double TYPICAL_ESCAPE_RANGE = 0.9;
    
    private List<DuelGun<T>> _guns;
    private Map<DuelGun<T>, Map<String, GunStats>> _gunRatings;

    public VirtualGunsManager() {
        _guns = new ArrayList<DuelGun<T>>();
        _gunRatings = new HashMap<DuelGun<T>, Map<String, GunStats>>();
    }

    public void addGun(DuelGun<T> gun) {
        _guns.add(gun);
        _gunRatings.put(gun, new HashMap<String, GunStats>());
    }

    public List<DuelGun<T>> getGuns() {
        return _guns;
    }

    public boolean contains(DuelGun<T> gun) {
        return _guns.contains(gun);
    }

    public double getRating(DuelGun<T> gun, String botName) {
        if (_guns.contains(gun) && _gunRatings.get(gun).containsKey(botName)) {
            return _gunRatings.get(gun).get(botName).gunRating();
        }

        return 0;
    }

    public int getShotsFired(DuelGun<T> gun, String botName) {
        if (_guns.contains(gun) && _gunRatings.get(gun).containsKey(botName)) {
            return _gunRatings.get(gun).get(botName).shotsFired;
        }

        return 0;
    }

    public void fireVirtualBullets(DiaWave w) {
        for (DuelGun<T> gun : _guns) {
            GunStats stats;
            if (_gunRatings.get(gun).containsKey(w.botName)) {
                stats = _gunRatings.get(gun).get(w.botName);
            } else {
                stats = new GunStats();
                _gunRatings.get(gun).put(w.botName, stats);
            }

            double firingAngle = gun.aimWithWave(w, false);
            stats.virtualBullets.put(w,
                new VirtualBullet(w.fireTime, firingAngle));
        }
    }

    public void registerWaveBreak(DiaWave w, double hitAngle,
            double tolerance) {
        for (DuelGun<T> gun : _guns) {
            GunStats stats = _gunRatings.get(gun).get(w.botName);
            VirtualBullet vb = stats.virtualBullets.get(w);

            double hitWeight = (w.targetDistance / TYPICAL_DISTANCE) *
                (w.escapeAngleRange() / TYPICAL_ESCAPE_RANGE);
            double missFactor = 
                Math.abs(Utils.normalRelativeAngle(vb.firingAngle - hitAngle))
                / tolerance;
            if (missFactor < 1) {
                stats.shotsHit += hitWeight * (1 - DiaUtils.square(missFactor));
            }

            stats.shotsFired++;
            stats.virtualBullets.remove(w);
        }
    }

    public void clear() {
        for (DuelGun<T> gun : _guns) {
            for (GunStats stats : _gunRatings.get(gun).values()) {
                stats.virtualBullets.clear();
            }
            gun.clear();
        }
    }

    public void clearWave(DiaWave w) {
        for (DuelGun<T> gun : _guns) {
            gun.clearWave(w);
        }
    }

    public DuelGun<T> bestGun(String botName) {
        DuelGun<T> bestGun = null;
        double bestRating = 0;

        for (DuelGun<T> gun : _guns) {
            double rating = 0;
            if (_gunRatings.get(gun).containsKey(botName)) {
                rating = _gunRatings.get(gun).get(botName).gunRating();
            }

            if (bestGun == null || rating > bestRating) {
                bestGun = gun;
                bestRating = rating;
            }
        }

        return bestGun;
    }

    public void printGunRatings(String botName) {
        System.out.println("Virtual Gun ratings for " + botName + ":");
        for (DuelGun<T> gun : _guns) {
            if (_gunRatings.get(gun).containsKey(botName)) {
                double rating = _gunRatings.get(gun).get(botName).gunRating();
                System.out.println("  " + gun.getLabel() + ": " + 
                    DiaUtils.round(rating * 100, 2));
            } else {
                System.out.println("WARNING (gun): Never logged any Virtual " +
                    "Guns info for " + gun.getLabel());
            }
        }
    }

    class GunStats {
        public int shotsFired;
        public double shotsHit;
        public HashMap<DiaWave, VirtualBullet> virtualBullets;

        public GunStats() {
            shotsFired = 0;
            shotsHit = 0;
            virtualBullets = new HashMap<DiaWave, VirtualBullet>();            
        }

        public double gunRating() {
            if (shotsFired == 0)
                return 0;

            return (shotsHit / shotsFired);
        }
    }

    class VirtualBullet {
        public final long fireTime;
        public final double firingAngle;

        public VirtualBullet(long fireTime, double firingAngle) {
            this.fireTime = fireTime;
            this.firingAngle = firingAngle;
        }
    }
}
