package voidious.move;

import java.awt.geom.Point2D;
import java.util.HashMap;
import java.util.Iterator;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import ags.utils.KdTree;

import robocode.Rules;
import voidious.utils.KnnView;
import voidious.utils.DiaUtils;
import voidious.utils.TimestampedGuessFactor;

/*
 * General info about the enemy. We just keep one of these around for 
 * each enemy bot the movement is aware of.
 * 
 * Copyright (c) 2009-2011 - Voidious
 *
 * This software is provided 'as-is', without any express or implied
 * warranty. In no event will the authors be held liable for any damages
 * arising from the use of this software.
 *
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, and to alter it and redistribute it
 * freely, subject to the following restrictions:
 *
 *    1. The origin of this software must not be misrepresented; you must not
 *    claim that you wrote the original software.
 *
 *    2. Altered source versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 *
 *    3. This notice may not be removed or altered from any source
 *    distribution.
 */

public class EnemyMoveData {
    private static final double RECENT_SCANS_HIT_THRESHOLD = 3;
    private static final double DECAY_RATE = 2;

    public String botName;
    public double distance;
    public double energy;
    public double velocity;
    public Point2D.Double location;
    public long lastScanTime;
    public ArrayList<Point2D.Double> pastLocations;
    public double heading;
    public double absBearing;
    public double damageTaken;
    public double damageGiven;
    boolean alive;
    public double lastBulletPower;
    public long lastBulletFireTime;
    public double totalBulletPower;
    public long totalTimesHit;
    public double totalDistance;
    public long timeAliveTogether;
    public long lastTimeHit;
    public long lastTimeClosest;
    private Map<String, Double> _botDistancesSq;
    public boolean flattenerEnabled;
    public boolean avoidBeingTargeted;
    public boolean stayPerpendicular;
    public double damageFactor;

    public double raw1v1ShotsFired;
    public double raw1v1ShotsHit;
    public double weighted1v1ShotsHit;
    public double raw1v1ShotsFiredThisRound;
    public double raw1v1ShotsHitThisRound;
    public double weighted1v1ShotsHitThisRound;
    public double lastRoundNormalized1v1HitPercentage;

    public List<KnnView<TimestampedGuessFactor>> views;
    public KdTree.WeightedSqrEuclid<Double> powerTree;

    public EnemyMoveData(String name, double dist, double en, Point2D.Double l, 
            double h, double ab, long lst) {
        botName = name;
        distance = dist;
        energy = en;
        velocity = 0;
        location = l;
        heading = h;
        absBearing = ab;
        damageTaken = 0;
        damageGiven = 0;
        alive = true;
        lastBulletPower = 0;
        totalBulletPower = 0;
        totalTimesHit = 0;
        totalDistance = 500;
        timeAliveTogether = 1;
        lastScanTime = lst;
        raw1v1ShotsFired = 0;
        weighted1v1ShotsHit = 0;
        flattenerEnabled = false;
        avoidBeingTargeted = false;
        stayPerpendicular = false;

        _botDistancesSq = new HashMap<String, Double>();
        pastLocations = new ArrayList<Point2D.Double>();

        initViews();
        powerTree = new KdTree.WeightedSqrEuclid<Double>(
            bulletPowerWeights.length, null);
        powerTree.setWeights(bulletPowerWeights);
    }

    public void initViews() {
        views = new ArrayList<KnnView<TimestampedGuessFactor>>();

        KnnView<TimestampedGuessFactor> simple =
            new KnnView<TimestampedGuessFactor>(new SimpleFormula())
                .setWeight(1)
                .setK(20)
                .setKDivisor(5)
                .bulletHitsOn();
        KnnView<TimestampedGuessFactor> simple2 =
            new KnnView<TimestampedGuessFactor>(new Simple2Formula())
                .setWeight(1)
                .setK(20)
                .setKDivisor(5)
                .bulletHitsOn();
        KnnView<TimestampedGuessFactor> simple3 =
            new KnnView<TimestampedGuessFactor>(new Simple3Formula())
                .setWeight(1)
                .setK(20)
                .setKDivisor(5)
                .bulletHitsOn();

        KnnView<TimestampedGuessFactor> normal =
            new KnnView<TimestampedGuessFactor>(new NormalFormula())
                .setWeight(40)
                .setK(20)
                .setKDivisor(5)
                .setHitThreshold(3.0)
                .bulletHitsOn();

        KnnView<TimestampedGuessFactor> recent =
            new KnnView<TimestampedGuessFactor>(new NormalFormula())
                .setWeight(100)
                .setK(1)
                .setMaxDataPoints(1)
                .setHitThreshold(RECENT_SCANS_HIT_THRESHOLD)
                .bulletHitsOn();
        KnnView<TimestampedGuessFactor> recent2 =
            new KnnView<TimestampedGuessFactor>(new NormalFormula())
                .setWeight(100)
                .setK(1)
                .setMaxDataPoints(5)
                .setHitThreshold(RECENT_SCANS_HIT_THRESHOLD)
                .bulletHitsOn();
        KnnView<TimestampedGuessFactor> recent3 =
            new KnnView<TimestampedGuessFactor>(new NormalFormula())
                .setWeight(100)
                .setK(1)
                .setHitThreshold(RECENT_SCANS_HIT_THRESHOLD)
                .setDecayRate(DECAY_RATE)
                .bulletHitsOn();
        KnnView<TimestampedGuessFactor> recent4 =
            new KnnView<TimestampedGuessFactor>(new NormalFormula())
                .setWeight(100)
                .setK(7)
                .setKDivisor(3)
                .setHitThreshold(RECENT_SCANS_HIT_THRESHOLD)
                .setDecayRate(DECAY_RATE)
                .bulletHitsOn();
        KnnView<TimestampedGuessFactor> recent5 =
            new KnnView<TimestampedGuessFactor>(new NormalFormula())
                .setWeight(100)
                .setK(35)
                .setKDivisor(3)
                .setHitThreshold(RECENT_SCANS_HIT_THRESHOLD)
                .setDecayRate(DECAY_RATE)
                .bulletHitsOn();

        KnnView<TimestampedGuessFactor> flattener =
            new KnnView<TimestampedGuessFactor>(new FlattenerFormula())
                .setWeight(50)
                .setK(25)
                .setMaxDataPoints(300)
                .setKDivisor(12)
                .flattenerOnly()
                .visitsOn();
        KnnView<TimestampedGuessFactor> flattener2 =
            new KnnView<TimestampedGuessFactor>(new FlattenerFormula())
                .setWeight(500)
                .setK(50)
                .setMaxDataPoints(2000)
                .setKDivisor(14)
                .flattenerOnly()
                .setDecayRate(DECAY_RATE)
                .visitsOn();

        views.add(simple);
        views.add(simple2);
        views.add(simple3);
        views.add(normal);
        views.add(recent);
        views.add(recent2);
        views.add(recent3);
        views.add(recent4);
        views.add(recent5);
        views.add(flattener);
        views.add(flattener2);
    }

    public double avgBulletPower() {
        if (totalBulletPower == 0) { return 3.0; }

        return totalBulletPower / totalTimesHit;
    }

    public void setBotDistanceSq(String name, double distance) {
        _botDistancesSq.put(name, distance);
    }

    public void clearDistancesSq() {
        _botDistancesSq.clear();
    }

    public void removeDistanceSq(String botName) {
        _botDistancesSq.remove(botName);
    }

    public double minDistanceSq() {
        double min = Double.POSITIVE_INFINITY;

        Iterator<Double> distIterator = _botDistancesSq.values().iterator();
        while (distIterator.hasNext()) {
            double dist = distIterator.next();
            if (dist < min) {
                min = dist;
            }
        }

        return min;
    }

    public int botsCloser(double distanceSq) {
        int bc = 0;

        Iterator<Double> distIterator = _botDistancesSq.values().iterator();
        while (distIterator.hasNext()) {
            double dist = distIterator.next();
            if (dist < distanceSq) {
                bc++;
            }
        }

        return bc;
    }

    public Point2D.Double getPastLocation(int x) {
        return pastLocations.get(pastLocations.size() - 1 - x);
    }

    public void clearNeighborCache() {
        for (KnnView<TimestampedGuessFactor> view : views) {
            view.clearCache();
        }
    }

    public double getGunHeat(long time) {
        double gunHeat;
        if (lastBulletPower == 0) {
            gunHeat = Math.max(0, 3.0 - (time * .1));
        } else {
            gunHeat = Math.max(0, Rules.getGunHeat(lastBulletPower) - 
                ((time - lastBulletFireTime) * .1));
        }

        return DiaUtils.round(gunHeat, 6);
    }

    private static double[] bulletPowerWeights = new double[]{3, 5, 1};
    protected double[] bulletPowerDataPoint(double distance, double enemyEnergy, 
        double myEnergy) {

        return new double[]{
            Math.min(distance, 800) / 800, 
            Math.min(enemyEnergy, 125) / 125,
            Math.min(myEnergy, 125) / 125
        };
    }
}
