package voidious.utils;

import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.awt.geom.Line2D;
import java.util.ArrayList;
import java.util.List;

import robocode.AdvancedRobot;
import robocode.util.Utils;

import voidious.utils.trace.TraceLogger;

/**
 * Copyright (c) 2009-2011 - Voidious
 *
 * This software is provided 'as-is', without any express or implied
 * warranty. In no event will the authors be held liable for any damages
 * arising from the use of this software.
 *
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, and to alter it and redistribute it
 * freely, subject to the following restrictions:
 *
 *    1. The origin of this software must not be misrepresented; you must not
 *    claim that you wrote the original software.
 *
 *    2. Altered source versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 *
 *    3. This notice may not be removed or altered from any source
 *    distribution.
 */

public class DiaUtils {
    public static boolean traceEnabled = false;

    private static final double HALF_PI = Math.PI / 2;
    private static final double TWO_PI = Math.PI * 2;
    private static final double BOT_HALF_WIDTH = 18;
    private static TraceLogger _traceLogger;

    public static Point2D.Double project(Point2D.Double sourceLocation, 
            double angle, double length) {
        return project(sourceLocation, Math.sin(angle), Math.cos(angle), length);
    }

    public static Point2D.Double project(Point2D.Double sourceLocation, 
            double sinAngle, double cosAngle, double length) {
        return new Point2D.Double(sourceLocation.x + sinAngle * length,
            sourceLocation.y + cosAngle * length);
    }

    public static double absoluteBearing(Point2D.Double sourceLocation,
           Point2D.Double target) {
        return Math.atan2(target.x - sourceLocation.x, 
            target.y - sourceLocation.y);
    }       

    public static int nonZeroSign(double d) {
        if (d < 0) { return -1; }
        return 1;
    }

    public static double square(double d) {
        return d * d;
    }

    public static double cube(double d) {
        return d * d * d;
    }

    public static double power(double d, int exp) {
        double r = 1;
        for (int x = 0; x < exp; x++) {
            r *= d;
        }
        return r;
    }

    public static double limit(double min, double value, double max) {
        return Math.max(min, Math.min(value, max));
    }

    public static int limit(int min, int value, int max) {
        return Math.max(min, Math.min(value, max));
    }

    public static double botWidthAimAngle(double distance) {
        return Math.abs(18.0/distance);
    }

    public static int bulletTicksFromPower(double distance, double power) {
        return (int)Math.ceil(distance / (20 - (3*power)));
    }

    public static int bulletTicksFromSpeed(double distance, double speed) {
        return (int)Math.ceil(distance / speed);
    }

    public static void setBackAsFront(AdvancedRobot robot, double goAngle) {
        double angle = Utils.normalRelativeAngle(
            goAngle - robot.getHeadingRadians());

        if (Math.abs(angle) > (HALF_PI)) {
            if (angle < 0) {
                robot.setTurnRightRadians(Math.PI + angle);
            } else {
                robot.setTurnLeftRadians(Math.PI - angle);
            }

            robot.setBack(100);
        } else {
            if (angle < 0) {
                robot.setTurnLeftRadians(-1 * angle);
           } else {
                robot.setTurnRightRadians(angle);
           }
            robot.setAhead(100);
        }
    }

    public static double rollingAverage(double previousValue, double newValue,
            double depth) {
        return ((previousValue * depth) + newValue) / (depth + 1);
    }

    public static double round(double d, int i) {
        long powerTen = 1;
        for (int x = 0; x < i; x++) {
            powerTen *= 10;
        }
        return ((double)Math.round(d * powerTen)) / powerTen;
    }

    public static double orbitalWallDistance(Point2D.Double sourceLocation,
            Point2D.Double targetLocation, double bulletPower, int direction, 
            BattleField battleField) {
        return orbitalWallDistance(sourceLocation, targetLocation, bulletPower,
            direction, battleField, 1);
    }

    public static double orbitalWallDistance(Point2D.Double sourceLocation,
            Point2D.Double targetLocation, double bulletPower, int direction, 
            BattleField battleField, double fudge) {
        double absBearing = 
            DiaUtils.absoluteBearing(sourceLocation, targetLocation);
        double distance = sourceLocation.distance(targetLocation) * fudge;
        double maxEscapeAngle = Math.asin(8.0/(20-3.0*bulletPower));

        // 1.0 means the max range of orbital movement exactly reaches bounds
        // of battle field
        double wallDistance = 2.0;
        for (int x = 0; x < 200; x++) {
            if (!battleField.rectangle.contains(
                sourceLocation.x + (Math.sin(absBearing +
                    (direction*(x/100.0)*maxEscapeAngle))*distance),
                sourceLocation.y + (Math.cos(absBearing +
                    (direction*(x/100.0)*maxEscapeAngle))*distance))) {
                wallDistance = x/100.0;
                break;
            }
        }

        return wallDistance;
    }

    public static double directToWallDistance(Point2D.Double targetLocation, 
            double distance, double heading, double bulletPower, 
            BattleField battleField) {
        int bulletTicks = DiaUtils.bulletTicksFromPower(distance, bulletPower);
        double wallDistance = 2.0;
        double sinHeading = Math.sin(heading);
        double cosHeading = Math.cos(heading);
        for (int x = 0; x < 2 * bulletTicks; x++) {
            if (!battleField.rectangle.contains(
                    targetLocation.x + (sinHeading * 8.0 * x),
                    targetLocation.y + (cosHeading * 8.0 * x))) {
                wallDistance = ((double)x)/bulletTicks;
                break;
            }
        }

        return wallDistance;
    }

    /**
     * wallSmoothing: do some Voodoo and wall smooth in a very efficient way.
     *   - ...in terms of CPU cycles, not amount of code.
     *   - used to be iterative, which was a lot simpler and more readable,
     *     but far too slow with how much it was called during precise
     *     prediction.
     */
    public static double wallSmoothing(BattleField battleField,
            Point2D.Double startLocation, double startAngle, int orientation,
            double wallStick) {
        double angle = startAngle;
        double wallDistanceX = Math.min(startLocation.x - 18, 
            battleField.width - startLocation.x - 18);
        double wallDistanceY = Math.min(startLocation.y - 18, 
            battleField.height - startLocation.y - 18);

        if (wallDistanceX > wallStick && wallDistanceY > wallStick) {
            return startAngle;
        }

        double testX = startLocation.x + (Math.sin(angle)*wallStick);
        double testY = startLocation.y + (Math.cos(angle)*wallStick);
        double testDistanceX = Math.min(testX - 18,
            battleField.width - testX - 18);
        double testDistanceY = Math.min(testY - 18,
            battleField.height - testY - 18);

        double adjacent = 0;
        int g = 0;

        while ((testDistanceX < 0 || testDistanceY < 0) && g++ < 25) {
            while (angle < 0) { angle += (2*Math.PI); }
            if (testDistanceY < 0 && testDistanceY < testDistanceX) {
                // wall smooth North or South wall
                angle = ((int)((angle + (Math.PI/2)) / Math.PI)) * Math.PI;
                adjacent = Math.abs(wallDistanceY);
            } else if (testDistanceX < 0 && testDistanceX <= testDistanceY) {
                // wall smooth East or West wall
                angle = (((int)(angle / Math.PI)) * Math.PI) + (Math.PI/2);
                adjacent = Math.abs(wallDistanceX);
            }

            angle += orientation *
                (Math.abs(Math.acos(adjacent/wallStick)) + 0.0005);

            testX = startLocation.x + (Math.sin(angle)*wallStick);
            testY = startLocation.y + (Math.cos(angle)*wallStick);
            testDistanceX = Math.min(testX - 18,
                battleField.width - testX - 18);
            testDistanceY = Math.min(testY - 18,
                battleField.height - testY - 18);
        }

        return angle;
    }

    public static double standardDeviation(double[] values) {
        double avg = average(values);
        double sumSquares = 0;
        for (int x = 0; x < values.length; x++) {
            sumSquares += square(avg - values[x]);
        }
        return Math.sqrt(sumSquares / values.length);
    }

    public static double average(double[] values) {
        double sum = 0;
        for (int x = 0; x < values.length; x++) {
            sum += values[x];
        }
        return (sum / values.length);
    }

    public static double accel(double velocity, double previousVelocity) {
        double accel = velocity - previousVelocity;
        // TODO: new decel-through-zero rules
        if (previousVelocity == 0.0) {
            accel = Math.abs(accel);
        } else {
            accel *= Math.signum(previousVelocity);
        }
        return accel;
    }

    public static Point2D.Double translateToField(Point2D.Double p, 
            BattleField battleField) {
        return new Point2D.Double(
            limit(18, p.x, battleField.width - 18), 
            limit(18, p.y, battleField.height - 18));
    }

    public static double preciseFrontBumperOffset(Point2D.Double sourceLocation,
            Point2D.Double botLocation) {
        return sourceLocation.distance(botLocation) - 
            distancePointToBot(sourceLocation, botLocation);
    }

    public static Rectangle2D.Double botRect(Point2D.Double botLocation) {
        Rectangle2D.Double botRect = new Rectangle2D.Double(botLocation.x - 18,
            botLocation.y - 18, 36, 36);
        return botRect;
    }

    public static List<Line2D.Double> botSides(Point2D.Double botLocation) {
        ArrayList<Line2D.Double> botSides = new ArrayList<Line2D.Double>();
        botSides.add(new Line2D.Double(botLocation.x - 18, 
            botLocation.y - 18, botLocation.x + 18, botLocation.y - 18));
        botSides.add(new Line2D.Double(botLocation.x + 18, 
            botLocation.y - 18, botLocation.x + 18, botLocation.y + 18));
        botSides.add(new Line2D.Double(botLocation.x + 18, 
            botLocation.y + 18, botLocation.x - 18, botLocation.y + 18));
        botSides.add(new Line2D.Double(botLocation.x - 18, 
            botLocation.y + 18, botLocation.x - 18, botLocation.y - 18));

        return botSides;
    }

    public static double distancePointToBot(
            Point2D.Double sourceLocation, Point2D.Double botLocation) {
        if (sourceLocation.x > botLocation.x - 18 &&
                sourceLocation.x < botLocation.x + 18 &&
                sourceLocation.y > botLocation.y - 18 &&
                sourceLocation.y < botLocation.y + 18) {
            return 0;
        } else {
            List<Line2D.Double> botSides = botSides(botLocation);
            double distance = Double.POSITIVE_INFINITY;
            for (Line2D.Double side : botSides) {
                distance = Math.min(distance, side.ptSegDist(sourceLocation));
            }
            return distance;
        }
    }

    public static List<Point2D.Double> botCorners(Point2D.Double p) {
        ArrayList<Point2D.Double> corners = new ArrayList<Point2D.Double>();
        corners.add(
            new Point2D.Double(p.x - BOT_HALF_WIDTH,p.y - BOT_HALF_WIDTH));
        corners.add(
            new Point2D.Double(p.x - BOT_HALF_WIDTH, p.y + BOT_HALF_WIDTH));
        corners.add(
            new Point2D.Double(p.x + BOT_HALF_WIDTH, p.y - BOT_HALF_WIDTH));
        corners.add(
            new Point2D.Double(p.x + BOT_HALF_WIDTH, p.y + BOT_HALF_WIDTH));

        return corners;
    }

    public static double normalizeAngle(double angle, double reference) {
        double normDiff = reference - angle;
        while (Math.abs(normDiff) > Math.PI) {
            angle += Math.signum(normDiff) * TWO_PI;
            normDiff = reference - angle;
        }
        return angle;
    }

    public static void log(String className, String methodName, String params,
            boolean entering) {
        if (traceEnabled) {
            _traceLogger.log(className, methodName, params, entering);
        }
    }

    public static void initTrace(String[] include, String[] exclude) {
        if (traceEnabled) {
            _traceLogger = new TraceLogger(include, exclude);
        }
    }

    public static double[] generateFiringAngles(int numAngles,
            double maxEscapeAngle) {
        int gfZero = (numAngles-1)/2;
        double[] firingAngles = new double[numAngles];
        for (int x = 0; x < numAngles; x++) {
            firingAngles[x] = (((double)(x - gfZero)) / gfZero) *
                maxEscapeAngle;
        }
        return firingAngles;
    }
}
