package voidious.utils;

import java.awt.geom.Line2D;
import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import robocode.util.Utils;
import voidious.utils.geom.Circle;
import voidious.utils.geom.LineSeg;

/**
 * Copyright (c) 2009-2011 - Voidious
 *
 * This software is provided 'as-is', without any express or implied
 * warranty. In no event will the authors be held liable for any damages
 * arising from the use of this software.
 *
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, and to alter it and redistribute it
 * freely, subject to the following restrictions:
 *
 *    1. The origin of this software must not be misrepresented; you must not
 *    claim that you wrote the original software.
 *
 *    2. Altered source versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 *
 *    3. This notice may not be removed or altered from any source
 *    distribution.
 */

public class DiaWave {
    public static final Point2D.Double ORIGIN = new Point2D.Double(0, 0);
    public static final double MAX_GUESSFACTOR = 1.0;
    public static final int CLOCKWISE = 1;
    public static final int COUNTERCLOCKWISE = -1;
    public static final boolean FIRING_WAVE = true;
    public static final boolean SURFABLE_WAVE = true;
    public static final boolean ANY_WAVE = false;
    public static final boolean POSITIVE_GUESSFACTOR = true;
    public static final boolean NEGATIVE_GUESSFACTOR = false;
    public static final double NO_CACHED_ESCAPE_ANGLE = -1;
    public static final double ANY_BULLET_POWER = -1;
    public static final int FIRST_WAVE = 0;
    protected static final double BOT_HALF_WIDTH = 18;
    protected static final double MAX_WAVE_INTERCEPT_OFFSET = 
        BOT_HALF_WIDTH / Math.cos(Math.PI / 4);
    protected static final double TWO_PI = Math.PI * 2;

    public String botName;
    public Point2D.Double sourceLocation;
    public Point2D.Double targetLocation;
    public double absBearing;
    public long fireTime;
    public double bulletPower;
    public double bulletSpeed;
    public double maxEscapeAngle;
    public int orbitDirection;
    public double targetHeading;
    public double targetRelativeHeading;
    public double targetVelocity;
    protected BattleField _battleField;

    public boolean processedBulletHit;
    public boolean processedPassed;
    public boolean processedWaveBreak;
    public boolean firingWave;
    public boolean altWave;

    public double targetAccel;
    public int targetVelocitySign;
    public double targetDistance;
    public double targetDistanceToNearestBot;
    public double targetDchangeTime;
    public double targetVchangeTime;
    public double targetWallDistance;
    public double targetRevWallDistance;
    public double targetDl8t;
    public double targetDl20t;
    public double targetDl40t;
    public double targetEnergy;
    public double sourceEnergy;
    public double gunHeat;
    public int enemiesAlive;
    public long lastBulletFiredTime;

    protected Point2D.Double waveBreakLocation;
    protected long waveBreakTime;
    public double waveBreakDistance;
    public Point2D.Double prevLocation;
    public long prevScanTime;

    public List<RobotState> waveBreakStates;
    public boolean interceptStarted;
    public List<BulletShadow> shadows;

    // used for data collection
    public int id;
    public int lastSeenId;

    protected double _cachedPositiveEscapeAngle = NO_CACHED_ESCAPE_ANGLE;
    protected double _cachedNegativeEscapeAngle = NO_CACHED_ESCAPE_ANGLE;
    public boolean usedNegativeSmoothingMea = false;
    public boolean usedPositiveSmoothingMea = false;

    public DiaWave() { }

    public DiaWave(String botName,
            Point2D.Double sourceLocation,
            Point2D.Double targetLocation, 
            long fireTime,
            double bulletPower,
            double targetHeading,
            double targetVelocity,
            int targetVelocitySign,
            BattleField battleField) {
        this.sourceLocation = sourceLocation;
        this.targetLocation = targetLocation;
        this.fireTime = fireTime;
        setBulletPower(bulletPower);
        absBearing = DiaUtils.absoluteBearing(sourceLocation, targetLocation);
        this.botName = botName;
        this.targetHeading = targetHeading;
        this.targetVelocity = targetVelocity;
        this.targetVelocitySign = targetVelocitySign;

        double relativeHeading = Utils.normalRelativeAngle(
            effectiveHeading() - 
                DiaUtils.absoluteBearing(sourceLocation, targetLocation));
        if (relativeHeading < 0) {
            orbitDirection = COUNTERCLOCKWISE;
        } else {
            orbitDirection = CLOCKWISE;
        }
        targetRelativeHeading = Math.abs(relativeHeading); 

        _battleField = battleField;
        processedBulletHit = false;
        processedPassed = false;
        processedWaveBreak = false;
        firingWave = false;
        altWave = false;
        waveBreakLocation = null;
        waveBreakStates = new ArrayList<RobotState>();
        interceptStarted = false;
        prevLocation = targetLocation;
        prevScanTime = fireTime;
        shadows = new ArrayList<BulletShadow>();
    }

    public void setBulletPower(double power) {
        bulletPower = power;
        bulletSpeed = (20 - (3 * power));
        maxEscapeAngle = Math.asin(8.0 / bulletSpeed);
    }

    public DiaWave setAccel(double accel) {
        this.targetAccel = accel;
        return this;
    }

    public DiaWave setDistance(double distance) {
        this.targetDistance = distance;
        return this;
    }

    public DiaWave setDistanceToNearestBot(double distanceToNearestBot) {
        this.targetDistanceToNearestBot = distanceToNearestBot;
        return this;
    }

    public DiaWave setDchangeTime(double dChangeTime) {
        this.targetDchangeTime = dChangeTime;
        return this;
    }

    public DiaWave setVchangeTime(double vChangeTime) {
        this.targetVchangeTime = vChangeTime;
        return this;
    }

    public DiaWave setWallDistance(double wallDistClock, double wallDistCounter) {
        if (orbitDirection == COUNTERCLOCKWISE && enemiesAlive == 1) {
            targetWallDistance = wallDistCounter;
            targetRevWallDistance = wallDistClock;
        } else {
            // 1v1 and orbiting clockwise 
            // OR melee and ignoring how they're orbiting
            targetWallDistance = wallDistClock;
            targetRevWallDistance = wallDistCounter;
        }
        return this;
    }

    public DiaWave setDistanceLast8Ticks(double dl8t) {
        this.targetDl8t = dl8t;
        return this;
    }

    public DiaWave setDistanceLast20Ticks(double dl20t) {
        this.targetDl20t = dl20t;
        return this;
    }

    public DiaWave setDistanceLast40Ticks(double dl40t) {
        this.targetDl40t = dl40t;
        return this;
    }

    public DiaWave setTargetEnergy(double energy) {
        this.targetEnergy = energy;
        return this;
    }

    public DiaWave setSourceEnergy(double energy) {
        this.sourceEnergy = energy;
        return this;
    }

    public DiaWave setGunHeat(double gunHeat) {
        this.gunHeat = gunHeat;
        return this;
    }

    public DiaWave setEnemiesAlive(int enemiesAlive) {
        this.enemiesAlive = enemiesAlive;
        return this;
    }

    public DiaWave setLastBulletFiredTime(long lastBulletFiredTime) {
        this.lastBulletFiredTime = lastBulletFiredTime;
        return this;
    }

    public DiaWave setAltWave(boolean altWave) {
        this.altWave = altWave;
        return this;
    }

    public boolean wavePassedInterpolate(Point2D.Double lastScanLocation, 
            long lastScanTime, long currentTime) {
        return wavePassedInterpolate(lastScanLocation, lastScanTime,
            currentTime, 0);
    }

    // Note: This method really needs to be called every tick for it to
    //       work correctly.
    public boolean wavePassedInterpolate(Point2D.Double lastScanLocation, 
            long lastScanTime, long currentTime, double offset) {
        if (processedPassed) {
            return true;
        } else if (lastScanTime == currentTime) {
            if (interceptStarted || 
                    pastFrontBumper(lastScanLocation, currentTime, offset)) {
                interceptStarted = true;
                if (currentTime - prevScanTime > 1) {
                    interpolateWaveBreakStates(lastScanLocation, currentTime,
                        offset);
                }

                if (processedPassed ||
                        waveGone(lastScanLocation, currentTime)) {
                    if (waveBreakStates.isEmpty()) {
                        logEmptyWaveBreakStates(currentTime, lastScanLocation);
                    }
                    processedPassed = true;
                } else {
                    waveBreakStates.add(new RobotState(lastScanLocation, 0, 0, 
                        lastScanTime));
                }
            }

            prevLocation = lastScanLocation;
            prevScanTime = lastScanTime;
        }

        return processedPassed;
    }

    private void interpolateWaveBreakStates(Point2D.Double lastScanLocation,
            long currentTime, double offset) {
        double deltaDistance = lastScanLocation.distance(prevLocation)
            / (currentTime - prevScanTime);
        double deltaBearing = DiaUtils.absoluteBearing(prevLocation,
            lastScanLocation);
        double dbSin = Math.sin(deltaBearing);
        double dbCos = Math.cos(deltaBearing);
        for (long x = prevScanTime + 1; x < currentTime; x++) {
            long interpoTime = x - prevScanTime;
            Point2D.Double interpoLocation = DiaUtils.project(prevLocation,
                dbSin, dbCos, (interpoTime) * deltaDistance);
            if (pastFrontBumper(interpoLocation, x, offset)) {
                if (waveGone(interpoLocation, x)) {
                    int medianBreakIndex = Math.min(
                        (waveBreakStates.size() / 2) +
                            (waveBreakStates.size() % 2),
                        waveBreakStates.size() - 1);
                    RobotState waveBreakState =
                        waveBreakStates.get(medianBreakIndex);

                    waveBreakLocation = waveBreakState.location;
                    waveBreakTime = waveBreakState.time;
                    processedPassed = true;
                    break;
                } else {
                    waveBreakStates.add(
                        new RobotState(interpoLocation, 0, 0, x));
                }
            }
        }
    }

    public boolean pastFrontBumper(Point2D.Double lastScanLocation,
            long currentTime, double offset) {
        double enemyDistSq = sourceLocation.distanceSq(lastScanLocation);
        double bulletDistance = bulletSpeed * (currentTime - fireTime + 1);
        if (enemyDistSq < DiaUtils.square(offset + bulletDistance +
                MAX_WAVE_INTERCEPT_OFFSET) &&
            enemyDistSq < DiaUtils.square(offset + bulletDistance + 
                DiaUtils.preciseFrontBumperOffset(sourceLocation, 
                    lastScanLocation))) {
            return true;
        } else {
            return false;
        }
    }

    public boolean wavePassed(Point2D.Double enemyLocation, long time,
            double offset) {
        double threshold = bulletSpeed * (time - fireTime) + offset;
        if (enemyLocation.distanceSq(sourceLocation) 
                < DiaUtils.square(threshold) * Math.signum(threshold)) {
            return true;
        } else {
            return false;
        }
    }

    public boolean waveGone(Point2D.Double enemyLocation, long time) {
        double bulletDistanceTraveledSq = DiaUtils.square(bulletSpeed * 
            (time - fireTime));
        List<Point2D.Double> corners = DiaUtils.botCorners(enemyLocation);
        for (Point2D.Double corner : corners) {
            if (corner.distanceSq(sourceLocation) > bulletDistanceTraveledSq) {
                return false;
            }
        }

        return true;
    }

    public Point2D.Double waveBreakLocation() {
        if (waveBreakLocation == null) {
            return prevLocation;
        } else {
            return waveBreakLocation;
        }
    }

    public long waveBreakTime() {
        if (waveBreakLocation == null) {
            return prevScanTime;
        } else {
            return waveBreakTime;
        }
    }

    public long waveBreakBulletTicks() {
        return waveBreakBulletTicks(waveBreakTime());
    }

    public long waveBreakBulletTicks(long waveBreakTime) {
        return waveBreakTime - fireTime;
    }

    public double effectiveHeading() {
        return Utils.normalAbsoluteAngle(targetHeading + 
            (targetVelocitySign == 1 ? 0 : Math.PI));
    }

    public Point2D.Double displacementVector() {
        return displacementVector(waveBreakLocation(), waveBreakTime());
    }

    public Point2D.Double displacementVector(Point2D.Double botLocation, 
            long time) {
        double vectorBearing = Utils.normalRelativeAngle(
            DiaUtils.absoluteBearing(targetLocation, botLocation) -
                effectiveHeading());
        double vectorDistance = targetLocation.distance(botLocation) /
            waveBreakBulletTicks(time);
        return DiaUtils.project(ORIGIN, vectorBearing * orbitDirection, 
            vectorDistance);
    }

    public double firingAngleFromDisplacementVector(Point2D.Double dispVector) {
        return Utils.normalAbsoluteAngle(DiaUtils.absoluteBearing(
            sourceLocation, projectLocation(dispVector)));
    }

    public double firingAngleFromTargetLocation(Point2D.Double firingTarget) {
        return Utils.normalAbsoluteAngle(DiaUtils.absoluteBearing(
            sourceLocation, firingTarget));
    }

    public Point2D.Double projectLocationBlind(Point2D.Double myNextLocation,
            Point2D.Double dispVector, long currentTime) {
        return projectLocation(myNextLocation, dispVector, 
            currentTime - fireTime + 1);
    }

    public Point2D.Double projectLocation(Point2D.Double dispVector) {
        return projectLocation(sourceLocation, dispVector, 0);
    }

    public Point2D.Double projectLocation(Point2D.Double firingLocation,
            Point2D.Double dispVector, long extraTicks) {
        double dispAngle = effectiveHeading() +  
            (DiaUtils.absoluteBearing(ORIGIN, dispVector) * orbitDirection);
        double dispDistance = ORIGIN.distance(dispVector);

        Point2D.Double projectedLocation;
        long bulletTicks = DiaUtils.bulletTicksFromSpeed(
            firingLocation.distance(targetLocation), bulletSpeed) - 1;
        long prevBulletTicks = 0;
        long prevPrevBulletTicks;
        int sanityCounter = 0;
        double daSin = Math.sin(dispAngle);
        double daCos = Math.cos(dispAngle);

        do {
            projectedLocation = DiaUtils.project(targetLocation,
                daSin, daCos, (bulletTicks + extraTicks) * dispDistance);
            prevPrevBulletTicks = prevBulletTicks;
            prevBulletTicks = bulletTicks;
            bulletTicks = DiaUtils.bulletTicksFromSpeed(
                firingLocation.distance(projectedLocation), bulletSpeed) - 1;

            if (bulletTicks == prevPrevBulletTicks) {
                projectedLocation = DiaUtils.project(targetLocation, daSin, daCos, 
                    (((bulletTicks + prevBulletTicks) / 2.0) + extraTicks) 
                        * dispDistance);
                break;
            }
        } while (bulletTicks != prevBulletTicks && sanityCounter++ < 20);

        return projectedLocation; 
    }

    public double distanceTraveled(long currentTime) {
        return (currentTime - fireTime) * bulletSpeed;
    }

    public double guessFactor(Point2D.Double targetLocation) {
        double bearingToTarget = 
            DiaUtils.absoluteBearing(sourceLocation, targetLocation);
        return guessFactor(bearingToTarget);
    }

    public double guessFactor(double bearingToTarget) {
        double guessAngle = orbitDirection * Utils.normalRelativeAngle( 
            bearingToTarget - absBearing);
        return guessAngle / maxEscapeAngle;
    }

    public double guessFactorPrecise(Point2D.Double targetLocation) {
        double newBearingToTarget = 
            DiaUtils.absoluteBearing(sourceLocation, targetLocation);
        return guessFactorPrecise(newBearingToTarget);
    }

    public double guessFactorPrecise(double newBearingToTarget) {
        double guessAngle = orbitDirection * Utils.normalRelativeAngle( 
                newBearingToTarget - absBearing);
        double maxEscapeAngle = preciseEscapeAngle(guessAngle >= 0);
        return guessAngle / maxEscapeAngle;
    }

    public double lateralVelocity() {
        return Math.sin(targetRelativeHeading) * Math.abs(targetVelocity);
    }

    public double escapeAngleRange() {
        return (preciseEscapeAngle(POSITIVE_GUESSFACTOR) 
                + preciseEscapeAngle(NEGATIVE_GUESSFACTOR));
    }

    public double preciseEscapeAngle(boolean guessFactorSign) {
        if (guessFactorSign) {
            if (_cachedPositiveEscapeAngle == NO_CACHED_ESCAPE_ANGLE) {
                _cachedPositiveEscapeAngle =
                    calculatePreciseEscapeAngle(guessFactorSign).angle 
                        * MAX_GUESSFACTOR;
            } 
            return _cachedPositiveEscapeAngle;
        } else {
            if (_cachedNegativeEscapeAngle == NO_CACHED_ESCAPE_ANGLE) {
                _cachedNegativeEscapeAngle =
                    calculatePreciseEscapeAngle(guessFactorSign).angle
                        * MAX_GUESSFACTOR;
            } 
            return _cachedNegativeEscapeAngle;
        }
    }

    public double firingAngle(double guessFactor) {
        return absBearing + 
            (guessFactor * orbitDirection * maxEscapeAngle);
    }

    public double virtuality() {
        long timeSinceLastBullet = (fireTime - lastBulletFiredTime);
        long timeToNextBullet = Math.round(Math.ceil(gunHeat * 10));

        if (firingWave) {
            return 0;
        } else if (lastBulletFiredTime > 0) {
            return Math.min(timeSinceLastBullet, timeToNextBullet) / 8.0;
        } else {
            return Math.min(1, timeToNextBullet / 8.0);
        }
    }

    public FiringTarget calculatePreciseEscapeAngle(
            boolean positiveGuessFactor) {
        return calculatePreciseEscapeAngle(positiveGuessFactor,
            new RobotState((Point2D.Double)targetLocation.clone(),
                targetHeading, targetVelocity, fireTime));
    }

    public FiringTarget calculatePreciseEscapeAngle(boolean positiveGuessFactor,
            RobotState startState) {
        boolean hitWall = false;
        boolean wavePassed = false;
        RobotState predictedState = (RobotState)startState.clone();

        boolean clockwisePrediction =
            (orbitDirection == 1 && positiveGuessFactor) ||
            (orbitDirection == -1 && !positiveGuessFactor);

        double noSmoothingEscapeAngle = 0;
        double bulletVelocity = this.bulletSpeed;

        do {    
            predictedState = Predictor.nextPerpendicularLocation(
                predictedState.location, absBearing, 
                predictedState.velocity, predictedState.heading, 
                clockwisePrediction, predictedState.time,
                Predictor.IGNORE_WALLS);

            if (!_battleField.rectangle.contains(predictedState.location)) {
                hitWall = true;
            } else if (
                wavePassed(predictedState.location, predictedState.time,
                    bulletVelocity)) {
                wavePassed = true;
            }
        } while (!hitWall && !wavePassed);

        noSmoothingEscapeAngle = 
            Math.abs(Utils.normalRelativeAngle(DiaUtils.absoluteBearing(
            sourceLocation, predictedState.location) - absBearing));
        FiringTarget noSmoothingTarget = new FiringTarget(
            noSmoothingEscapeAngle, predictedState.location,
            predictedState.time);

        FiringTarget bestSmoothingTarget = null;
        if (hitWall) {
            double wallSmoothingStick = 80;
            double purelyPerpendicularAttackAngle = 0;
            double fullVelocity = 8.0;
            double orbitAbsBearing = absBearing;
            bestSmoothingTarget = new FiringTarget(0, startState.location,
                startState.time);

            for (int x = 0; x < 3; x++) {
                wavePassed = false;
                predictedState = new RobotState(
                    (Point2D.Double)targetLocation.clone(), targetHeading,
                    targetVelocity, fireTime);

                do {    
                    predictedState = 
                        Predictor.nextPerpendicularWallSmoothedLocation(
                            predictedState.location, orbitAbsBearing, 
                            predictedState.velocity, fullVelocity, 
                            predictedState.heading, 
                            purelyPerpendicularAttackAngle, clockwisePrediction,
                            predictedState.time, _battleField, 
                            wallSmoothingStick, Predictor.OBSERVE_WALL_HITS);
                    if (wavePassed(predictedState.location,
                            predictedState.time, bulletVelocity)) {
                        wavePassed = true;
                    }                   
                } while (!wavePassed);

                orbitAbsBearing = DiaUtils.absoluteBearing(targetLocation, 
                    predictedState.location) -
                    ((clockwisePrediction ? 1 : -1) * (Math.PI / 2));

                double thisSmoothingEscapeAngle =
                    Math.abs(Utils.normalRelativeAngle(DiaUtils.absoluteBearing(
                    sourceLocation, predictedState.location) - absBearing));
                if (thisSmoothingEscapeAngle > bestSmoothingTarget.angle) {
                    bestSmoothingTarget = new FiringTarget(
                        thisSmoothingEscapeAngle, predictedState.location,
                        predictedState.time);
                }
            }
        }

        if (bestSmoothingTarget != null && 
                bestSmoothingTarget.angle > noSmoothingTarget.angle) {
            return bestSmoothingTarget;
        } else {
            return noSmoothingTarget;
        }
    }

    public void clearCachedPreciseEscapeAngles() {
        _cachedPositiveEscapeAngle = NO_CACHED_ESCAPE_ANGLE;
        _cachedNegativeEscapeAngle = NO_CACHED_ESCAPE_ANGLE;
    }

    public boolean shadowed(double firingAngle) {
        for (BulletShadow shadow : shadows) {
            firingAngle = DiaUtils.normalizeAngle(firingAngle, shadow.minAngle);
            if (firingAngle >= shadow.minAngle && 
                    firingAngle <= shadow.maxAngle) {
                return true;
            }
        }
        return false;
    }

    public void castShadow(Point2D.Double p1, Point2D.Double p2) {
        double shadowAngle1 = DiaUtils.normalizeAngle(
            DiaUtils.absoluteBearing(sourceLocation, p1), absBearing);
        double shadowAngle2 = DiaUtils.normalizeAngle(
            DiaUtils.absoluteBearing(sourceLocation, p2),
            absBearing);
        double min = Math.min(shadowAngle1, shadowAngle2);
        double max = Math.max(shadowAngle1, shadowAngle2);
        boolean merged = false;
        for (BulletShadow shadow : shadows) {
            if (min > shadow.minAngle && min < shadow.maxAngle) {
                max = shadow.maxAngle = Math.max(max, shadow.maxAngle);
                merged = true;
            } else if (max > shadow.minAngle && max < shadow.maxAngle) {
                min = shadow.minAngle = Math.min(min, shadow.minAngle);
                merged = true;
            } else if (min < shadow.minAngle && max > shadow.maxAngle) {
                shadow.minAngle = min;
                shadow.maxAngle = max;
                merged = true;
            }
        }

        if (!merged) {
            shadows.add(new BulletShadow(min, max));
        }
    }

    public double shadowFactor(double angle, double bandwidth) {
        double min = angle - bandwidth;
        double max = angle + bandwidth;
        double factor = 1;
        for (BulletShadow shadow : shadows) {
            double shadowMin = DiaUtils.normalizeAngle(shadow.minAngle, min);
            double shadowMax = DiaUtils.normalizeAngle(shadow.maxAngle, min);
            if (shadowMin < min && shadowMax > max) {
                return 0;
            } else if (shadowMin > min && shadowMin < max) {
                factor -= (Math.min(max, shadowMax) - shadowMin) / (max - min);
            } else if (shadowMax > min && shadowMax < max) {
                factor -= (shadowMax - Math.max(min, shadowMin)) / (max - min);
            }
        }

        return factor;
    }

    public static class BulletShadow {
        public double minAngle;
        public double maxAngle;

        public BulletShadow(double minAngle, double maxAngle) {
            this.minAngle = minAngle;
            this.maxAngle = maxAngle;
        }
    }

    private void logEmptyWaveBreakStates(long time,
            Point2D.Double lastScanLocation) {
        String botCorners = "Bot corners:";
        for (Point2D.Double corner : DiaUtils.botCorners(lastScanLocation)) {
            botCorners += "\n  " + DiaUtils.round(corner.x, 1)
                + ", " + DiaUtils.round(corner.y, 1);
        }
        
        String error =
            "Anomaly in wave break interpolation - no wave break states.\n"
            + "bulletSpeed: " + bulletSpeed + "\n"
            + "sourceLocation: " + DiaUtils.round(sourceLocation.x, 1)
            + ", " + DiaUtils.round(sourceLocation.y, 1) + "\n"
            + "fireTime: " + fireTime + "\n"
            + "lastScanLocation: " + DiaUtils.round(lastScanLocation.x, 1)
                + ", " + DiaUtils.round(lastScanLocation.y, 1) + "\n"
            + "lastScanTime: currentTime = " + time + "\n"
            + botCorners + "\n"
            + "prevLocation: " + DiaUtils.round(prevLocation.x, 1)
                + ", " + DiaUtils.round(prevLocation.y, 1) + "\n"
            + "prevScanTime: " + prevScanTime + "\n"
            + "previous bumper offset: " + DiaUtils.round(
                DiaUtils.preciseFrontBumperOffset(sourceLocation, prevLocation),
                    4);
        ErrorLogger.logError(error);
    }

    public static class FiringTarget {
        public double angle;
        public Point2D.Double location;
        public long time;

        public FiringTarget(double angle, Point2D.Double location, long time) {
            this.angle = angle;
            this.location = location;
            this.time = time;
        }
    }

    public static DiaWave findSurfableWave(List<DiaWave> waves, int waveIndex,
            Point2D.Double targetLocation, long currentTime) {
        // TODO: sort by time to impact?
        int searchWaveIndex = 0;
        DiaWave surfableWave = null;

        Iterator<DiaWave> wavesIterator = waves.iterator();
        while (wavesIterator.hasNext()) {
            DiaWave w = wavesIterator.next();
            double distanceToWaveSource =
                targetLocation.distance(w.sourceLocation);
            double distanceToWave = distanceToWaveSource -
                w.distanceTraveled(currentTime);

            if (!w.processedBulletHit && distanceToWave > w.bulletSpeed &&
                    w.firingWave) {
                if (searchWaveIndex == waveIndex) {
                    surfableWave = w;
                    break;
                } else {
                    searchWaveIndex++;
                }
            }
        }
        
        return surfableWave;
    }

    public static DiaWave findClosestWave(List<DiaWave> waves,
            Point2D.Double targetLocation, long currentTime,
            double matchThreshold, double bulletPower) {
        return findClosestWave(waves, targetLocation, currentTime,
            DiaWave.ANY_WAVE, matchThreshold, null, bulletPower);
    }

    public static DiaWave findClosestWave(List<DiaWave> waves,
            Point2D.Double targetLocation, long currentTime,
            boolean onlyFiring, double matchThreshold, String botName,
            double bulletPower) {
        double closestDistance = Double.POSITIVE_INFINITY;
        DiaWave closestWave = null;

        for (DiaWave w : waves) {
            if (!w.altWave && (botName == null || botName.equals(w.botName) || 
                botName.equals(""))) {
                double distanceFromTargetToWave =
                    (w.sourceLocation.distance(targetLocation) -
                        w.distanceTraveled(currentTime));

                if (Math.abs(distanceFromTargetToWave) < matchThreshold &&
                        Math.abs(distanceFromTargetToWave) < closestDistance &&
                        (!onlyFiring || w.firingWave) &&
                        (bulletPower == DiaWave.ANY_BULLET_POWER ||
                         Math.abs(bulletPower - w.bulletPower) < 0.001)) {
                    closestDistance = Math.abs(distanceFromTargetToWave);
                    closestWave = w;
                }
            }
        }

        return closestWave;     
    }

    public double[] preciseBotWidth(List<RobotState> dangerStates) {
        if (dangerStates == null || dangerStates.size() == 0) {
            return null;
        }

        ArrayList<Double> aimAngles = new ArrayList<Double>();
        for (RobotState dangerState : dangerStates) {
            List<Point2D.Double> corners =
                DiaUtils.botCorners(dangerState.location);
            Circle waveStart = new Circle(sourceLocation.x, sourceLocation.y, 
                bulletSpeed * (dangerState.time - fireTime));
            Circle waveEnd = new Circle(sourceLocation.x, sourceLocation.y,
                bulletSpeed * (dangerState.time - fireTime + 1));

            for (Point2D.Double corner : corners) {
                if (waveEnd.contains(corner) && !waveStart.contains(corner)) {
                    aimAngles.add(DiaUtils.absoluteBearing(sourceLocation, 
                        corner));
                }
            }

            for (Line2D.Double side : dangerState.botSides()) {
                LineSeg seg = new LineSeg(side.x1, side.y1, side.x2, side.y2);
                Point2D.Double[] intersects = waveStart.intersects(seg);
                for (int x = 0; x < intersects.length; x++) {
                    if (intersects[x] != null) {
                        aimAngles.add(DiaUtils.absoluteBearing(sourceLocation, 
                            intersects[x]));
                    }
                }
                intersects = waveEnd.intersects(seg);
                if (intersects != null) {
                    for (int x = 0; x < intersects.length; x++) {
                        if (intersects[x] != null) {
                            aimAngles.add(DiaUtils.absoluteBearing(
                                sourceLocation, intersects[x]));
                        }
                    }
                }
            }
        }

        if (aimAngles.size() == 0) {
            // Caused when the second wave is past any danger zone by the
            // time we finish surfing the first. This can be caused by
            // chase bullets or the like. Sorting waves by time to impact
            // instead of distance in 1.5.10 mostly eliminated this
            // (tested vs PrairieWolf), but lost points. Reverted.
            return new double[]{
                DiaUtils.absoluteBearing(sourceLocation, 
                    dangerStates.get(0).location),
                DiaUtils.botWidthAimAngle(
                    dangerStates.get(0).location.distance(sourceLocation))
                        / Math.asin(8.0 / bulletSpeed)};
        }

        double normalizeReference = aimAngles.get(0);
        double minAngle = normalizeReference;
        double maxAngle = normalizeReference;

        for (double thisAngle : aimAngles) {
            // TODO: switch to normalizeAngle method
            double normDiff = normalizeReference - thisAngle;
            while (Math.abs(normDiff) > Math.PI) {
                thisAngle += Math.signum(normDiff) * TWO_PI;
                normDiff = normalizeReference - thisAngle;
            }

            if (thisAngle > maxAngle) {
                maxAngle = thisAngle;
            }

            if (thisAngle < minAngle) {
                minAngle = thisAngle;
            }
        }

        double centerAngle = (maxAngle + minAngle) / 2;
        double bandwidth = maxAngle - centerAngle;

        return new double[]{centerAngle, bandwidth};
    }
}
