package xander.cat;

import java.awt.Color;
import java.awt.geom.Path2D;
import java.util.Map;

import robocode.WinEvent;

import xander.cat.drive.AutoVaryWeightByAge;
import xander.cat.drive.GreenBeeSurfSelector;
import xander.cat.drive.HighPerformanceDriveSelector;
import xander.cat.drive.IdealPositionDrive;
import xander.cat.drive.SmartDistancingEquation;
import xander.cat.group.mirror.MirrorFactory;
import xander.cat.group.ram.RamFactory;
import xander.cat.group.rem.REMFactory;
import xander.cat.gun.BSProtectedGun;
import xander.cat.gun.power.SteppedHitRatioPowerSelector;
import xander.cat.gun.targeter.JPCircularTargeter;
import xander.cat.processor.TargetingDetectorFactorArrayProcessor;
import xander.cat.radar.BasicRadar;
import xander.cat.scenario.CircularDriveScenario;
import xander.cat.scenario.NoOpponentWavesScenario;
import xander.core.AbstractXanderRobot;
import xander.core.ComponentChain;
import xander.core.Configuration;
import xander.core.Resources;
import xander.core.RobotStyle;
import xander.core.Scenario;
import xander.core.drive.Drive;
import xander.core.drive.DriveArray;
import xander.core.drive.DriveBoundsFactory;
import xander.core.drive.DriveSelector;
import xander.core.gun.Gun;
import xander.core.gun.GunArray;
import xander.core.gun.VirtualHitRatioGunSelector;
import xander.core.gun.XanderGun;
import xander.core.gun.detect.HeadOnDetector;
import xander.core.gun.detect.LinearDetector;
import xander.core.gun.detect.TargeterTargetingDetector;
import xander.core.gun.power.PowerSelector;
import xander.core.gun.targeter.CircularTargeter;
import xander.core.gun.targeter.Targeter;
import xander.core.log.Logger;
import xander.core.math.RCPhysics;
import xander.core.track.GunStats;
import xander.gfws.data.DataPointFactorArrayProcessor;
import xander.gfws.data.FactorArrayParms;
import xander.gfws.data.KDTreeWaveLogger;
import xander.gfws.data.KNNLogReader;
import xander.gfws.data.KNNParms;
import xander.gfws.data.RedKDTreeWaveLogger;
import xander.gfws.drive.DirectWaveSurfingDrive;
import xander.gfws.gun.targeter.GuessFactorTargeter;
import xander.gfws.segment.AttackerBearingSegmenter;
import xander.gfws.segment.BulletTravelTimeSegmenter;
import xander.gfws.segment.DefenderAccelerationSegmenter;
import xander.gfws.segment.DefenderSpeedSegmenter;
import xander.gfws.segment.LateralVelocitySegmenter;
import xander.gfws.segment.WallStickSegmenter;

/**
 * Robocode advanced robot and flagship of the Xander robot fleet.
 * XanderCat is a multi-mode robot built on the Xander 2.0 framework.
 * 
 * @author Scott Arnold
 */
public class XanderCat extends AbstractXanderRobot {

	private static DataPointFactorArrayProcessor driveFAP;
	private static CircularDriveScenario circularDriverScenario;
	private static SteppedHitRatioPowerSelector steppedPowerSelector;
	private static PowerSelector mainPowerSelector;
	private static Path2D.Double driveBounds;
	private static boolean[] wins; 
	
	@Override
	protected void style(RobotStyle robotStyle) {
		robotStyle.setColors(Color.WHITE, Color.BLACK, Color.GREEN);
	}

	@Override
	protected void configure(Configuration configuration) {
		configuration.setLogComponentRunTimes(true);
		configuration.setLogDriveTimes(true);
		driveBounds = DriveBoundsFactory.getSmoothedRectangleBounds(getBattleFieldSize());
		Path2D.Double oppDriveBounds = DriveBoundsFactory.getRectangularBounds(getBattleFieldSize());
		configuration.setUsePreciseMEAForOpponentWaves(true, oppDriveBounds);
		PowerSelector powerSelector = configuration.getDisabledOpponentPowerSelector();
		powerSelector = REMFactory.getX5PowerSelector(powerSelector);
		configuration.setDisabledOpponentPowerSelector(powerSelector);
		steppedPowerSelector = new SteppedHitRatioPowerSelector(
				new double[] {0.17}, new double[] {1.85, 1.99});
		steppedPowerSelector.setPowerDrop(0.135, 15, 0);
		mainPowerSelector = REMFactory.getX5PowerSelector(steppedPowerSelector);
		wins = new boolean[getNumRounds()]; 
	}

	private double getAveragedStatValue(String statKey, double newValue, Map<String, String> oppStats) {
		String statValue = oppStats.get(statKey);
		if (statValue != null) {
			try {
				double oldValue = Double.parseDouble(statValue);
				newValue = (newValue + oldValue) / 2d;
			} catch (Exception e) { }
		}
		return newValue;
	}
	
	@Override
	protected boolean recordBattleStats(Map<String, String> oppStats) {
		GunStats gunStats = Resources.getGunStats();
		double oHR = gunStats.getOverallOpponentHitRatio();
		oHR = getAveragedStatValue("OppHitRatio", oHR, oppStats);
		double mHR = gunStats.getOverallHitRatio();
		mHR = getAveragedStatValue("MyHitRatio", mHR, oppStats);
		double fop = driveFAP.getFlattenerUsePercentage();
		fop = getAveragedStatValue("FlattenerPercent", fop, oppStats);
		oppStats.put("OppHitRatio", Logger.format(oHR,3));
		oppStats.put("MyHitRatio", Logger.format(mHR,3));
		oppStats.put("FlattenerPercent", Logger.format(fop,3));
		for (int i=0; i<2; i++) {
			int bf = gunStats.getActionedBulletsFired("GF Gun " + i);
			oppStats.put("BF"+i, String.valueOf(bf));
		}
		double cda = circularDriverScenario.getAppliesPercentage();
		cda = getAveragedStatValue("CircularApplies", cda, oppStats);
		oppStats.put("CircularApplies", Logger.format(cda,3));
		double dpc = steppedPowerSelector.getDropPowerCount();
		dpc = getAveragedStatValue("DropPowerCount", dpc, oppStats);
		oppStats.put("DropPowerCount", Logger.format(dpc, 1));
		return true;
	}

	private int getIntValue(Map<String, String> stats, String key, int defaultValue) {
		String s = stats.get(key);
		if (s != null) {
			try {
				return Integer.parseInt(s);
			} catch (Exception e) { };
		}
		return defaultValue;
	}
	
	@Override
	protected boolean recordCommonBattleStats(Map<String, String> commonStats) {
		int totalBattles = getIntValue(commonStats, "Total Battles", 0);
		commonStats.put("Total Battles", String.valueOf(++totalBattles));
		for (int i=0; i<wins.length; i++) {
			String roundKey = (i < 10)? "R 0" + i + " Wins" : "R " + i + " Wins";
			int roundWins = getIntValue(commonStats, roundKey, 0);
			if (wins[i]) {
				roundWins++;
			}
			commonStats.put(roundKey, String.valueOf(roundWins));
		}
		return true;
	}

	@Override
	protected void addComponents(ComponentChain chain) {
		
//		FactorArrayPainter wsDriveFAPainter = new FactorArrayPainter("WS Drive FA", "WS Drive", false);
//		FactorArrayPainter tdDriveFAPainter = new FactorArrayPainter("TD Drive FA", "TD Drive", false);
//		tdDriveFAPainter.setPegColor(Color.PINK);
//		XanderPaintManager.getInstance(this).enable(getBattleFieldHeight(), wsDriveFAPainter, tdDriveFAPainter);
		
		// RADAR
		
		chain.addDefaultComponents(new BasicRadar(45, 5));
		
		// DRIVES
		
		RamFactory.addAntiRamComponents(chain);
		
		MirrorFactory.addAntiMirrorComponents(chain, 12);

		Scenario ipScenario = new NoOpponentWavesScenario();
		Drive ipDrive = new IdealPositionDrive();
		chain.addComponents(ipScenario, ipDrive);
		
		SmartDistancingEquation de = new SmartDistancingEquation(80, 400, 600, 45, 20);
		de.setMyHitRatioThresholds(0, 0);
		KNNParms driveKNNParms = new KNNParms(1, 30, 0.3);	
		KDTreeWaveLogger driveWaveLogger = new RedKDTreeWaveLogger(false,
				new LateralVelocitySegmenter(12),
				new DefenderAccelerationSegmenter(),
				new AttackerBearingSegmenter(16),
				new WallStickSegmenter(4, 60, 260, true));	
		FactorArrayParms driveFactorArrayParms = new FactorArrayParms(false); 
		driveFactorArrayParms.setUseFlattener(true, 0.125, 0.33, 5);
		driveFactorArrayParms.setMinAgeWeightPercent(0.06);
		new AutoVaryWeightByAge(driveFactorArrayParms, 0.09, 0.11, 0.0005, 0.1275, 0.005);
		KNNLogReader driveLogReader = new KNNLogReader(driveWaveLogger, driveKNNParms);			
		driveFAP = new DataPointFactorArrayProcessor("WS Drive FA", driveLogReader, 67, driveFactorArrayParms);
		GreenBeeSurfSelector surfSelector = new GreenBeeSurfSelector(driveFAP, de);
		DirectWaveSurfingDrive mainDrive = new DirectWaveSurfingDrive("WS Drive", surfSelector, driveBounds);
		mainDrive.setSurfMultipleWaves(true);	
		
		TargetingDetectorFactorArrayProcessor tdFAP = new TargetingDetectorFactorArrayProcessor("TD Drive FA", 67,
				new HeadOnDetector(false),
				new LinearDetector(false),
				new TargeterTargetingDetector(false, new JPCircularTargeter(1), RCPhysics.ROBOT_WIDTH),
				new TargeterTargetingDetector(false, new JPCircularTargeter(2), RCPhysics.ROBOT_WIDTH));
		GreenBeeSurfSelector tdSurfSelector = new GreenBeeSurfSelector(tdFAP, de);
		DirectWaveSurfingDrive tdDrive = new DirectWaveSurfingDrive("TD Drive", tdSurfSelector, driveBounds);
		tdDrive.setSurfMultipleWaves(true);
		
		DriveSelector driveSelector = new HighPerformanceDriveSelector(mainDrive, 50, 0.08);
		DriveArray driveArray = new DriveArray(driveSelector, mainDrive, tdDrive);		
						
		chain.addDefaultComponents(driveArray);
		
		// GUNS  
		
		// a special scenario just for our circular drivers out there!
		XanderGun circularGun = new XanderGun(new CircularTargeter(), mainPowerSelector);
		circularDriverScenario = new CircularDriveScenario(circularGun);
		chain.addComponents(circularDriverScenario, circularGun);
		
		// main guess factor guns
		KDTreeWaveLogger gunWaveLogger = new RedKDTreeWaveLogger(true,
				new DefenderSpeedSegmenter(16),
				new BulletTravelTimeSegmenter(24, getBattleFieldSize()),
				new DefenderAccelerationSegmenter(),
				new AttackerBearingSegmenter(24),
				new WallStickSegmenter(4, 60, 260, true),
				new WallStickSegmenter(4, 60, 260, false));	
		KNNParms gunKNNParms = new KNNParms(20, 100, 0.05);
		gunKNNParms.setKnnWhenSelectionEmpty(0);
		KNNLogReader gunLogReader = new KNNLogReader(gunWaveLogger, gunKNNParms);
		double gunRoll = 0;
		Gun[] guns = new Gun[2];
		for (int i=0; i<guns.length; i++) {
			FactorArrayParms gunFactorArrayParms = new FactorArrayParms(true);
			gunFactorArrayParms.setVaryWeightByAge(gunRoll);
			DataPointFactorArrayProcessor gunFAP = new DataPointFactorArrayProcessor("Gun FA", gunLogReader, 91, gunFactorArrayParms);
			Targeter targeter = new GuessFactorTargeter(gunFAP);
			XanderGun gun = new XanderGun("GF Gun " + i, targeter, mainPowerSelector);
			gun.setMinEnergyToFire(4.1d);
			gun.setLowEnergyConservationRate(1d/3d);
			guns[i] = new BSProtectedGun(gun);
			gunRoll += 0.005;
		}
		VirtualHitRatioGunSelector gunSelector = new VirtualHitRatioGunSelector();
		gunSelector.setRollingRatioWeight(0.25);
		GunArray gunArray = new GunArray(gunSelector, guns);
		
		
		chain.addDefaultComponents(gunArray);
	}

	@Override
	public void onWin(WinEvent event) {
		wins[getRoundNum()] = true;
		super.onWin(event);
	}		
}
