package xander.cat.group.mirror;

import java.awt.geom.Rectangle2D;

import robocode.ScannedRobotEvent;

import xander.core.Resources;
import xander.core.RobotEvents;
import xander.core.event.RoundBeginListener;
import xander.core.event.ScannedRobotListener;
import xander.core.log.Log;
import xander.core.log.Logger;
import xander.core.math.RCPhysics;
import xander.core.track.Snapshot;
import xander.core.track.SnapshotHistory;

/**
 * Detects opponents that attempt to perfectly mirror our driving pattern.
 * Presently only guaranteed to work properly for 1v1.
 * 
 * @author Scott Arnold
 */
public class MirrorDetector implements RoundBeginListener, ScannedRobotListener {

	private static final Log log = Logger.getLog(MirrorDetector.class);
	private static final double BOX_SIZE = RCPhysics.ROBOT_WIDTH + 10;
	
	private boolean[] mirrorHits;
	private Rectangle2D.Double mirrorBounds;
	private Rectangle2D.Double battlefieldBounds;
	private SnapshotHistory snapshotHistory;
	private int scanIndex;
	private double mirrorThreshold = 0.75;
	
	public MirrorDetector(int scanDepth) {
		this.snapshotHistory = Resources.getSnapshotHistory();
		if (scanDepth > snapshotHistory.getHistorySize()) {
			scanDepth = snapshotHistory.getHistorySize();
			log.warn("Robot histories are not long enough for requested scan depth; reducing scan depth to " + scanDepth);
		}
		this.mirrorHits = new boolean[scanDepth];
		this.battlefieldBounds = Resources.getRobotProxy().getBattleFieldSize();
		this.mirrorBounds = new Rectangle2D.Double(0, 0, 1, 1);  // initialize it to anything
		RobotEvents robotEvents = Resources.getRobotEvents();
		robotEvents.addScannedRobotListener(this);
	}

	@Override
	public void onRoundBegin() {
		// it will take the opponent a short while to get into mirror position, so reset mirror hits on startup
		for (int i=0; i<mirrorHits.length; i++) {
			mirrorHits[i] = false;
		}		
	}

	public void onScannedRobot(ScannedRobotEvent event) {
		// each scan we check for new histories
		// we don't worry about if it's the same robot for melee as it won't really cause any negative effects
		// opponent is going to lag behind a little, so grab my snapshot from a couple ticks ago
		Snapshot me = snapshotHistory.getMySnapshot(Math.max(0, event.getTime()-2));
		Snapshot opp = snapshotHistory.getSnapshot(event.getName(), event.getTime());
		if (me != null && opp != null) {
			double x = battlefieldBounds.getMaxX()-me.getX()-BOX_SIZE/2d;
			double y = battlefieldBounds.getMaxY()-me.getY()-BOX_SIZE/2d;
			mirrorBounds.setRect(x, y, BOX_SIZE, BOX_SIZE);
			mirrorHits[scanIndex] = mirrorBounds.contains(opp.getLocation());
		}
		// update scan index
		scanIndex++;
		if (scanIndex == mirrorHits.length) {
			scanIndex = 0;
		}
	}
	
	public boolean isMirrorDetected() {
		int hitCount = 0;
		for (boolean mirrorHit : mirrorHits) {
			if (mirrorHit) {
				hitCount++;
			}
		}
		float hitPercent = (float) hitCount / (float) mirrorHits.length;
		log.debug("Mirror percentage: " + Logger.format(100*hitPercent));
		return hitPercent > mirrorThreshold;
	}
}
